'''
Define a timeseries sequence.
'''

from .Sequence import *

class Timeseries_Sequence(Sequence):
    '''
    NOT CURRENTLY BEING USED.
    FIXME -- add support for updating the points in a plot in time.
    '''

    def __init__(self, initial, y=None, yuncertainty=None, name='timeseries', **kwargs):
        '''
        Initialize a Sequence from some 1D timeseries.

        Parameters
        ----------
        initial : LightCruve object, or array of times (either astropy times, or in JD)
                The time values, to be plotted on the x-axis
        y : array
                The values to be plotted on the y-axis
        yuncertainty : array
                The errorbars on y (optional).
        **kwargs are passed to plt.plot()
        '''

        try:
            # is it a lightkurve?
            time = initial.time
            y = initial.flux
        except:
            time = initial

        # make sure we have a TPF as imput
        # set up the basic sequence
        if isinstance(time, Time):
            self.time = time
        else:
            self.time = Time(
                time, format=guess_time_format(time), scale=timescale)

        # store the dependent values
        self.y = y
        self.yuncertainty = yuncertainty

        # create a sequence out of that stamp
        Sequence.__init__(self, name=name)

        # set a rough title for this plot
        self.titlefordisplay = '{}'.format(self.name)

    def __getitem__(self, timestep):
        '''
        Return the image data for a given timestep.
        '''
        if timestep is None:
            return None
        else:
            return self.y[timestep, :, :]



from .FrameBase import *
from ..sequences import make_image_sequence


class EmptyTimeseriesFrame(FrameBase):
    '''
    An EmptyTimeseriesFrame has a vertical bar for time when animated.

    The frame should be added to an illustration,
    the illustration should be plotted,
    and then data can be plotted into the
    ETF.ax and ETF.ax_hist frames.
    '''

    frametype = 'timeseries'

    def __init__(self, name='timeseries', xlim=[None, None], ylim=[None, None], ylabel='', histogram=True, **kwargs):
        '''
        Initialize an empty timeseries frame.

        Parameters
        ----------

        name : string
                A custom name for this frame.

        xlim : tuple
                Set the xlimits. If None, wait for data to appear.

        ylim : tuple
                Set the ylimits. If None, wait for data to appear.

        ylabel : string
                Set a ylabel for this plot.

        histogram : bool
                Should we add an extra axis for a histogram?
        '''
        # initialize an empty frame
        FrameBase.__init__(self, name=name, **kwargs)

        # store the plot limits
        self.xmin, self.xmax = xlim
        self.ymin, self.ymax = ylim

        # store the custom title for this one
        self.titlefordisplay = ylabel

        # should this frame have a histrogram attached to it?
        self.histogram = histogram

    def plot(self, t=None, y=None, **kwargs):
        '''
        Make an empty plot, with a vertical line waiting in it.
        '''

        self.plotted = {}

        if t is not None:
            self.plotted['data'] = self.ax.plot(t-self.offset, y, **kwargs)

        if self.histogram:
            b = self.ax.get_position()
            width = b.x1 - b.x0
            scale = 0.1
            self.ax_hist = self.ax.figure.add_axes(
                [b.x1, b.y0, b.width * scale, b.height])
            self.ax_hist.set_facecolor('lightgray')
            self.ax_hist.set_axis_off()
            self.ax.set_position([b.x0, b.y0, b.width, b.height])

        # plot the time bar
        self.plotted['vline'] = self.ax.axvline(
            np.nan, color='gray', alpha=0.5, zorder=-1e6)

        # pull the title for this frame
        self.ax.set_ylabel(self.titlefordisplay)
        if self.ax.get_xticklabels()[0].get_visible():
            self.ax.set_xlabel('Time - {:.5f} (days)'.format(self.offset))

        # turn the axes lines off
        # plt.axis('off')

        # keep track of the current plotted timestep
        self.currenttimestep = None

        # only set the limits if we need to; otherwise, let the first plot do it
        if self.xmin is not None and self.xmax is not None:
            plt.xlim(self.xmin, self.xmax)
        if self.ymin is not None and self.ymax is not None:
            plt.ylim(self.ymin, self.ymax)

    def update(self, time):
        '''
        Update this frame to a particular time (for use in animations).
        '''

        # add a vertical line
        v = time.jd - self.offset
        self.plotted['vline'].set_xdata(v)
        #self.speak('offset is {}'.format(self.offset))
        #self.speak('drew vertical line at {}'.format(v))
        # add a time label
        #s = self._timestring(time)
        # try:
        #	self.plotted['time'].set_text(s)
        # except KeyError:
        #	self.plotted['time'] = self.ax.text(0.01, +0.02, s, va='bottom', zorder=1e6, transform=self.ax.transAxes)

        self.currenttimestep = time



from ..imports import *


class FrameBase(Talker):

    frametype = 'base'
    timeunit = 'day'
    aspectratio = 1

    def __init__(self,
                    name='',
                    ax=None,
                    data=None,
                    illustration=None,
                    plotingredients=[],
                    **kwargs):
        '''
        Initialize this Frame, which can be one (of multiple)
        frames in an Illustration. Before plotting, each frame
        needs to have an `ax` defined saying where it should be
        plotted and (generally) some `data` associated with it.

        Parameters
        ----------

        name : str
            A name to give this Frame.

        ax : matplotlib.axes.Axes instance
            All plotting will happen inside this ax.
            If set to None, the `self.ax attribute` will
            need to be set manually before plotting.

        data : flexible
            It can be custom object (e.g. a Stamp),
            or simple arrays, or a list of HDULists,
            or something else, depending on
            what this actual Frame does with it
            in `plot` and `update`.

        illustration : an Illustration
            The illustration in which this frame will be embedded.
            We keep track of this (sometimes) because we might want
            information that's shared across the whole illustration
            to be accessible to this individual frame when plotting.

        plotingredients : list
            A list of keywords indicating features that will be
            plotted in this frame. It can be modified either now
            when initializing the frame, or any time before
            calling `i.plot()` from the illustration.

        **kwargs go nowhere (?)
        '''

        # initialize the Talker
        Talker.__init__(self, prefixformat='{:>32}')

        # store a name for this frame
        self.name = name

        # assign this frame an axes to sit in
        self.ax = ax

        # this is likely a Sequence of some kind
        self.data = data

        # is there another overarching frame this one should be aware of?
        self.illustration = illustration

        # what ingredients should be included when plotting?
        self.plotingredients = list(plotingredients)

        # keep track of a list of frames included in this one (e.g. zooms)
        self.includes = []

        # keep track of what's been plotted
        self.plotted = {}

        # KLUDGE for now -- need better time label management
        self.timeformat = 't={offset:.5f}{time:+.5f}'

    @property
    def offset(self):
        '''
        Get a time offset, to use as a zero-point.

        Returns
        -------
        offset : float
                An appropriate time zero-point (in JD).
        '''

        # is an offset already defined?
        try:
            return self._offset
        except AttributeError:
            # is there a minimum of the illustration's times?
            try:
                self._offset = np.min(self.illustration._get_times().jd)
            except (AttributeError, ValueError):
                # otherwise, use only this frame to estimate the offset
                try:
                    self._offset = np.min(self._get_times().jd)
                except ValueError:
                    self._offset = 0.0
            self.speak('defining {} as the time offset'.format(self._offset))

            return self._offset

    def _timestring(self, time):
        '''
        Return a string, given an input time.

        Parameters
        ----------
        time : astropy Time
                A particular time.

        Returns
        -------
        timestring : str
                A string describing the times.
        '''

        if self.data._timeisfake:
            timestep = self._find_timestep(time)
            return '#{:.0f}'.format(self._get_times()[timestep].gps)
        else:
            days = time.jd - self.offset
            inunits = (days * u.day).to(self.timeunit)

            return self.timeformat.format(offset=self.offset, time=inunits)

    def __repr__(self):
        '''
        Default string representation for this frame.
        '''
        return '<{} Frame | data={} | name={}>'.format(self.frametype,
                                                       self.data,
                                                       self.name)
    def __str__(self):
        return '<F"{}">'.format(self.name)

    def plot(self):
        '''
        This should be redefined in a class that inherits from FrameBase.
        '''
        raise RuntimeError(
            "Don't know how to `plot` {}".format(self.frametype))

    def update(self, *args, **kwargs):
        '''
        This should be redefined in a class that inherits from FrameBase.
        '''
        raise RuntimeError(
            "Don't know how to `update` {}".format(self.frametype))

    def _find_timestep(self, time):
        '''
        Given a time, identify its index.

        Parameters
        ----------

        time : float
                A single time (in JD?).

        Returns
        -------
        index : int
                The index of the *closest* time point.
        '''
        return self.data._find_timestep(time)

    def _get_times(self):
        '''
        Get the available times associated with this frame.
        '''

        # does the data have a time axis defined?
        try:
            return self.data.time
        # if not, return an empty time
        except AttributeError:
            return Time([], format='gps')

    def _timesandcadence(self, round=None):
        '''
        Get all the unique times available across all the frames,
        along with a suggested cadence set by the minimum
        (rounded) differences between times.

        Parameters
        ----------
        round : float
                All times will be rounded to this value.
                Times separated by less than this value
                will be considered identical.
        '''

        # store this calculation with the illustration, so it doesn't need repeating
        try:
            self._precaculatedtimesandcadence
        except AttributeError:
            self._precaculatedtimesandcadence = {}
        try:
            times, cadence = self._precaculatedtimesandcadence[round]
        except KeyError:
            gps = self._get_times().gps

            if round is None:
                diffs = np.diff(np.sort(gps))
                round = np.min(diffs[diffs > 0])

            baseline = np.min(gps)
            rounded = round * np.round((gps - baseline) / round) + baseline
            uniquegpstimes = np.unique(rounded)
            cadence = np.min(np.diff(uniquegpstimes)) * u.s
            times = Time(uniquegpstimes, format='gps')
            self._precaculatedtimesandcadence[round] = times, cadence

        return times, cadence

    def _transformimage(self, image):
        '''
        Some frames will want to flip or rotate an image before display.
        This handles that transformation. (This should probably be set
        up as an matplotlib.axes transform type of thing.)
        '''
        return image

    def _transformxy(self, x, y):
        '''
        This handles the same transformation as that which goes into
        transform image, but for x and y arrays.
        '''
        return x, y

    def _get_orientation(self):
        '''
        Figure out the orientation of the overarching illustration.
        '''
        try:
            return self.illustration.orientation
        except:
            return 'horizontal' #'vertical'



from .imshowFrame import *
from .CameraFrame import *

class CCDFrame(CameraFrame):
    '''
    This is an imshowFrame associated with one CCD. It can
    transform from CCD coordinates to Camera coordinates.

    A single CCD can be created on its own, but more often
    it will be linked to a Camera, to continue the transformation
    onward from Camera to display.
    '''

    frametype = 'CCD'

    def __init__(self, name='ccd',
                       ax=None,
                       data=None,
                       camera=CameraFrame(),
                       **kwargs):

        # if this CCD is associated with a named camera, label in this frame
        if camera.name != 'camera':
            name = '{}-{}'.format(camera.name, name)

        # initialize the base imshowFrame, with data and a name
        imshowFrame.__init__(self, name=name, ax=ax, data=data, **kwargs)

        # make sure we're linked to a camera (for transformations)
        self.camera = camera

        # set up the sizes
        self.xmin, self.ymin = 0, 0
        try:
            # if there's an image, use it to set the size
            self.ymax, self.xmax = self.data[0].shape
        except (IndexError, AttributeError, TypeError):
            # allow us to create an empty image
            self.ymax = 2078
            self.xmax = 2136

        # set the default translation for this CCD
        self.transpose = False
        self.flipy     = False
        self.flipx     = False

        # give a non-empty string title for this CCD
        # self.titlefordisplay = self.name

    def _transformimage(self, image):
        '''
        This first transforms by the CCD's transformation to get to
        Camera coordinates, and then transforms from Camera coordinates
        to get to display coordinates.

        It reuses the definition of CameraFrame's transformations,
        but first with the (transpose,flipx,flipy) of the CCD,
        and then after with the (transpose,flipx,flipy) of the Camera.
        '''
        cameraimage = CameraFrame._transformimage(self, image)
        return self.camera._transformimage(cameraimage)

    def _transformxy(self, x, y):
        '''
        This first transforms by the CCD's transformation to get to
        Camera coordinates, and then transforms from Camera coordinates
        to get to display coordinates.

        It reuses the definition of CameraFrame's transformations,
        but first with the (transpose,flipx,flipy) of the CCD,
        and then after with the (transpose,flipx,flipy) of the Camera.
        '''

        # shrink the camera to CCD size, so we can still use it for transformations (?)
        self.camera.xmax, self.camera.ymax = self.xmax, self.ymax
        camerax, cameray = CameraFrame._transformxy(self, x, y)
        return self.camera._transformxy(camerax, cameray)

class CCD1Frame(CCDFrame):
    '''
    CCDs 1 and 2 have (0,0) in the upper right.
    '''
    def __init__(self, name='ccd1', **kwargs):
        CCDFrame.__init__(self, name=name, **kwargs)
        self.transpose = False
        self.flipy     = True
        self.flipx     = True

class CCD2Frame(CCD1Frame):
    '''
    CCDs 1 and 2 have (0,0) in the upper right.
    '''
    def __init__(self, name='ccd2', **kwargs):
        CCD1Frame.__init__(self, name=name, **kwargs)

class CCD3Frame(CCDFrame):
    '''
    CCDs 3 and 4 have (0,0) in the lower left (untransformed).
    '''
    def __init__(self, name='ccd3', **kwargs):
        CCDFrame.__init__(self, name=name, **kwargs)
        self.transpose = False
        self.flipy     = False
        self.flipx     = False

class CCD4Frame(CCD3Frame):
    '''
    CCDs 3 and 4 have (0,0) in the lower left (untransformed).
    '''
    def __init__(self, name='ccd4', **kwargs):
        CCD3Frame.__init__(self, name=name, **kwargs)

# define a dictionary, so these ccds are easier to access
ccds = {'ccd' : CCDFrame,
        'ccd1': CCD1Frame,
        'ccd2': CCD2Frame,
        'ccd3': CCD3Frame,
        'ccd4': CCD4Frame}



from .stamps import *

def establish_file_lists(sparse_pattern, stamps_directory='stamps', nstars=None):
	print('Searching for "{}"...'.format(sparse_pattern))

	# write a complete list of all the files
	total_filename = os.path.join(stamps_directory, 'everysinglefile.txt')
	if os.path.exists(total_filename):
		with open(total_filename, 'r') as total:
			total_files = total.read().splitlines()
		print('{} lists {} sparse subarray files total.'.format(total_filename, len(total_files)))
	else:
		total_files = glob.glob(sparse_pattern)
		print('Found {} sparse subarray files.'.format(len(total_files)))
		with open(total_filename, 'w') as total:
			total.writelines([name + '\n' for name in total_files])

	completed_filename = os.path.join(stamps_directory, 'completedfiles.txt')
	if os.path.exists(completed_filename):
		with open(completed_filename, 'r') as complete:
			completed_files = complete.read().splitlines()
	else:
		completed_files = []
	print("{} lists {} completed sparse subarray files.".format(completed_filename, len(completed_files)))

	remaining_files = list(total_files)
	for f in completed_files:
		remaining_files.remove(f)
	print("{} files remain to split!".format(len(remaining_files)))

	return remaining_files



def split_times_to_stars(sparse_pattern, stamps_directory='stamps', ntimes=None, nstars=None):
	'''
	Take a group of sparse subarray FITS files,
	and split them up into directories for stars,
	each filled with individual time points.

	These can then be reassembled into Stamps.

	Parameters
	----------

	sparse_pattern : str
		A search string pattern for sparse subarray files.

	stamps_directory : str
		Directory into which all the stamps will be stored (organized by camera).

	ntimes : int
		How many times should be included? (defaults to all)

	nstars : int
		How many stars should be included? (defaults to all)

	Returns
	-------

	directories_pattern : str
		A search string pattern that will capture all the directories just created.
	'''

	# make sure the stamps directory exists
	mkdir(stamps_directory)
	for cam in [1,2,3,4]:
		mkdir(os.path.join(stamps_directory, 'cam{}'.format(cam)))


	# make a list of sparse subarray files
	sparse_files = establish_file_lists(sparse_pattern, stamps_directory)#glob.glob(sparse_pattern)[:ntimes]
	completed_filename = os.path.join(stamps_directory, 'completedfiles.txt')

	# loop over the files (times)
	for i, f in enumerate(sparse_files):

		print(f)
		# open this file (time)
		hdus = fits.open(f, memmap=False)[:nstars]

		# this is the whole frame
		frame = hdus[0]

		# loop over extensions (aka stars)
		for e in np.arange(0, len(hdus)):

			# this is the one specific star
			star = hdus[e]

			# create the static dictionary for this star
			static = {}

			# store the items that are static to this star
			for key in ['SPM', 'CAM', 'INT_TIME']:
				static[key] = frame.header[key]
			for key in ['TIC_ID', 'COL_CENT', 'ROW_CENT']:
				static[key] = star.header[key]

			temporal = {}
			for key in ['QUAL_BIT', 'TIME', 'CADENCE']:
				temporal[key] = frame.header[key]

			# where should the timestamps be stored
			directory = os.path.join(stamps_directory, 'cam{CAM}/cam{CAM}_spm{SPM}_tic{TIC_ID}'.format(**static))

			# save the static information once
			mkdir(directory)
			static_filename = os.path.join(directory, 'static.npy')
			#print('saved static information to {}'.format(static_filename))
			np.save(static_filename, static)

			# save the individual timestamp for this star
			timestamp_filename = os.path.join(directory, 'img-{:08}.npy'.format(i))
			np.save(timestamp_filename, (star.data, temporal))

		print('saved {} stars from file {}/{} at {} \r'.format(len(hdus), i+1, len(sparse_files), Time.now().iso))
		with open(completed_filename, 'a') as complete:
				completed_files = complete.write(f + '\n')
	return os.path.join(stamps_directory, 'cam*/cam*_spm*_tic*/')

def combine_times_to_stamps(directories_pattern, stamps_directory='stamps', ntimes=None, nstars=None):
	'''
	Convert a group of directories into Stamps.

	Parameters
	----------
	directories_pattern : str
		A search string for directories that contain a static.npy, and lots of img-*.npy files.

	ntimes : int
		How many times should be included? (defaults to all)

	nstars : int
		How many stars should be included? (defaults to all)

	'''
	stamps_directories = glob.glob(directories_pattern)[:nstars]

	# pull up all the stamp directories
	for star, d in enumerate(stamps_directories):
		print('   {} = {}'.format(star, d))
		print('')
		static_file = os.path.join(d, 'static.npy')
		static = np.load(static_file)[()]
		img_files = glob.glob(os.path.join(d, 'img-*.npy'))[:ntimes]

		N = len(img_files)
		for i, f in enumerate(img_files):
			#print(i, f)
			image, temp = np.load(f)
			if i == 0:
				temporal = dict(**temp)
				for k in temporal.keys():
					temporal[k] = np.empty(N).astype(type(temporal[k]))
				photons = np.zeros((N, image.shape[0], image.shape[1]))
			for k in temporal.keys():
				temporal[k][i] = temp[k]
			photons[i,:,:] = image
		print('{}/{}'.format(star, len(stamps_directories)))
		s = Stamp(photons=photons, static=static, temporal=temporal)
		directory = os.path.join(stamps_directory, 'cam{CAM}'.format(**static))
		s.save(s.filename(directory))

def process_sparse_into_stamps(sparse_pattern, stamps_directory, **kw):
	'''
	Construct a bunch of Stamps from a bunch of sparse subarray files.

	Parameters
	----------

	sparse_pattern : str
		A search string pattern for sparse subarray files.

	stamps_directory : str
		Directory into which all the stamps will be stored (organized by camera).

	ntimes : int
		How many times should be included? (defaults to all)

	nstars : int
		How many stars should be included? (defaults to all)
	'''
	directories_pattern = split_times_to_stars(sparse_pattern, stamps_directory=stamps_directory, **kw)
	combine_times_to_stamps(directories_pattern, stamps_directory=stamps_directory, **kw)



'''
Define some filename parsers than can pull details from filenames,
and convert them to dictionaries.
'''

from ..imports import *



def generic_filenameparser(filename):
    '''
    Parse a general file like, for example,
    "some-little-filename.fits"
    into a dictionary of important information.

    (This one is meant to be expanded upon elsewhere.)

    Parameters
    ----------
    filename : str
        The filename of a single image.

    Returns
    -------
    features : dict
        This dictionary of details contains the filename, (and you
        should add some other keywords too, if you want to expand
        upon this.)
    '''
    d = {}
    d['filename'] = os.path.basename(filename)
    return d

def explicit_filenameparser(filename):
    '''
    Parse a file that explicitly specifies ccd and camera
    like, for example, "cam1-ccd2-00001.fits"
    into a dictionary of important information.

    (This one is meant to be expanded upon elsewhere.)

    Parameters
    ----------
    filename : str
        The filename of a single image.

    Returns
    -------
    features : dict
        This dictionary of details contains the filename, camera, and CCD.
    '''
    d = {}
    d['filename'] = os.path.basename(filename)
    s = d['filename'].split('.fit')[0]
    try:
        d['camera'] = s[s.index('cam'):][3]
    except ValueError:
        pass
    try:
        d['ccd'] = s[s.index('ccd'):][3]
    except ValueError:
        pass
    return d

def tessqlp_filenameparser(filename):
    '''
    Parse a QLP-type (single-CCD) file like, for example,
    "tess2018220104526-00004907-1-crm-ffi-ccd1.fits"
    into a dictionary of important information.

    Parameters
    ----------
    filename : str
        The filename of a single FITS image.

    Returns
    -------
    features : dict
        This dictionary of details contains the filename, the cadence number,
        the camera number, whether or not crm was applied, the image type, and
        the CCD number.
    '''

    d = {}
    d['filename'] = os.path.basename(filename)
    s = d['filename'].split('.fit')[0]
    components = s.split('-')
    d['cadence'] = float(components[1])
    d['camera'] = int(components[2])
    d['crm'] = components[3]
    d['type'] = components[4]
    d['ccd'] = s[s.index('ccd'):][3]
    return d

def tessqlp_fullcamera_filenameparser(filename):
    '''
    Parse a QLP-type file like, for example,
    "tess2018220104526-00005279-4-crm-ffi_dehoc.fits"
    into a dictionary of important information.

    Parameters
    ----------
    filename : str
        The filename of a single FITS image.

    Returns
    -------
    features : dict
        This dictionary of details contains the filename, the cadence number,
        the camera number, whether or not crm was applied, and the image type.
    '''

    d = {}
    d['filename'] = os.path.basename(filename)
    s = d['filename'].split('.fit')[0]
    components = s.split('-')
    d['cadence'] = float(components[1])
    d['camera'] = int(components[2])
    d['crm'] = components[3]
    d['type'] = components[4]
    return d


def mast_filenameparser(filename):
    '''
    Parse a MAST-type (single-CCD) file like, for example,
    "tess2018206192942-s0001-1-1-0120-s_ffic.fits"
    into a dictionary of important information.

    Parameters
    ----------
    filename : str
        The filename of a single FITS image.

    Returns
    -------
    features : dict
        This dictionary of details contains the filename, the datetime string,
        an astropytime, a JD, the camera, the CCD, whether or not crm was applied,
        and the image type.
    '''

    d = {}
    d['filename'] = os.path.basename(filename)
    s = d['filename'].split('.fit')[0]
    components = s.split('-')

    d['datetime'] = components[0][4:]
    dt=d['datetime']
    year, day, hour, minute, second = dt[0:4], dt[4:7], dt[7:9], dt[9:11], dt[11:13]
    yday = '{}:{}:{}:{}:{}'.format(year, day, hour, minute, second)
    d['astropytime'] = Time(yday, format='yday')
    d['jd'] =  d['astropytime'].jd
    d['camera'] = int(components[2])
    d['ccd'] = int(components[3])
    d['crm'] = components[5][0] == 's'
    d['type'] = components[5][2:6]
    return d

def flexible_filenameparser(filename):
    '''
    Parse a generic FITS filename,
    trying to guess an appropriate method for it,
    and create a dictionary of important information.

    Parameters
    ----------
    filename : str
        The filename of a single FITS image.

    Returns
    -------
    features : dict
        This dictionary of details contains at least the filename, and
        hopefully some other stuff too.
    '''

    # filename parsers, in order of priority
    parsers = [ mast_filenameparser,
                tessqlp_filenameparser,
                tessqlp_fullcamera_filenameparser,
                explicit_filenameparser,
                generic_filenameparser]

    # try each parser in turn, moving to the next if one complains
    for p in parsers:
        try:
            return p(filename)
        except:
            pass



'''
These are tools for playing with and visualizing photometry
of TESS postage stamps. They operate exclusively on TPF files.
'''

from .processing import *
from .visualize import *
from .cosmics import *



from .Sequence import *
from .FITS_Sequence import *
from .Stamp_Sequence import *
from .TPF_Sequence import *
from .Timeseries_Sequence import *
from .filenameparsers import *

def make_image_sequence(initial, *args, **kwargs):
    '''
    Initialize a Sequence for viewing with tv.

    This is a general wrapper that *tries* to
    guess what kind of Sequence to make,
    based solely on the inputs. If this fails,
    you may need to specify the Sequence explicitly.

    Parameters
    -----------

        initial : (many possible types)
                This is the input that initializes an image sequence.
                    - single FITS filename, and an extension to use.
                    - list of FITS filenames, and an extension to use.
                    - a glob pattern to search for FITS files (and extension)
                    - single FITS HDUList, and an extension to use.
                    - list of loaded FITS HDULists, and an extension to use.
                    - a Stamp object from the `cosmics` package.

        *args
            Positional arguments will be passed on to whatever Sequence is initialized

        **kwargs
            Keyword arguments will be passed on to whatever Sequence is initialized

    '''

    # is it already a sequence?
    if issubclass(initial.__class__, Sequence):
        return initial
    # is it empty? return an empty sequence
    elif initial is [] or initial is None:
        return Sequence()
    # is it a Stamp?
    elif isinstance(initial, Cube):
        return Stamp_Sequence(initial, *args, **kwargs)
    elif type(initial) == np.ndarray:
        return Image_Sequence(initial, **kwargs)
    else:
        # try:
        #	# is initial a 1D thing?
        #	assert(len(np.shape(initial))==1 or isinstance(initial, LightCurve))
        #	return Timeseries_Sequence(initial, *args, **kwargs)
        # except AssertionError:

        # is it a TPF (or can it be used to make one, like a filename)?
        try:
            assert(isinstance(initial, TargetPixelFile))
            return TPF_Sequence(initial, *args, **kwargs)
        # if nothing else, assume it is a FITS_Sequence
        except (AttributeError, AssertionError):
            return FITS_Sequence(initial, *args, **kwargs)



'''
Define a generic sequence.
'''

from ..imports import *
from ..postage.stamps import Stamp
from ..utilities import *
try:
    from ..postage.tpf import EarlyTessTargetPixelFile
    from lightkurve.lightcurve import LightCurve
    from lightkurve.targetpixelfile import TargetPixelFile
except:
    raise RuntimeWarning(
        "no lightkurve found; some TPF tools may not be available")

timescale = 'tdb'
# by default, assume all times are TDB
# this is ahead of TAI by about 31.184s
# and it differs from UTC by additional leap seconds





class Sequence(Talker):
    '''
    A sequence of one or more images, which can be viewed (and animated) in a tv frame.
    '''

    titlefordisplay = ''
    colorbarlabelfordisplay = ''
    _timeisfake = False

    def __init__(self, name='generic', *args, **kwargs):
        Talker.__init__(self, prefixformat='{:>32}')
        self.name = name

    def cadence(self):
        '''

        Returns
        -------
        cadence : astropy Time
                A typical cadence for the sequence.
        '''

        return np.median(np.diff(self.time))

    def _find_timestep(self, time):
        '''
        Given a time, identify its index.

        Parameters
        ----------

        time : float
                A single time (in JD?).

        Returns
        -------
        index : int
                The index of the *closest* time point.
        '''

        # pull out actual times associated with this sequence
        times = self._get_times()

        # if there are no times, return nothing
        if len(times) == 0:
            return None
        else:
            # find the index of the timepoint that is closest to this one
            diff = time - times
            step = np.argmin(abs(diff))
            return step

    def _get_times(self):
        '''
        Get the available times associated with this frame.

        Returns
        -------
        times : array
                All of the times associated with this sequence.
        '''
        return self.time

    def __repr__(self):
        '''
        How should this sequence be represented, by default, as a string.
        '''
        return '<{} of {} elements>'.format(self.nametag, self.N)

    @property
    def N(self):
        '''
        How many elements are in this sequence?
        '''
        try:
            return len(self.time)
        except AttributeError:
            return 0

    def __len__(self):
        return self.N



'''
Define a sequence of images from a Stamp object.
'''
from .Image_Sequence import *
from ..postage.stamps import *

class Stamp_Sequence(Image_Sequence):
    def __init__(self, initial, name='Stamp', **kwargs):
        '''
        Initialize a Sequence from a Stamp.
        (a Stamp is a custom, simplified version of a TPF)

        Parameters
        ----------
        initial : (many possible types)
                -a single Stamp object
                -something that can initialize a Stamp
                        A filename for a '.npy' file.
                        A list of FITS sparse_subarray files.
                        A search path of FITS sparse_subarray files (e.g. containing *.fits)

        **kwargs : dict
                Keyword arguments are passed to Stamp(initial, **kwargs)
        '''

        # make sure we have at least a stamp
        if isinstance(initial, Stamp) or isinstance(initial, Cube):
            stamp = initial
        elif (type(initial) == str) or (type(initial) == list):
            stamp = Stamp(initial, **kwargs)


        # create a sequence out of that stamp
        Sequence.__init__(self, name=name)

        # (not coincidentally), a Stamp has similar variables to a Sequence
        for k in stamp._savable:
            vars(self)[k] = vars(stamp)[k]

        # set up the basic sequence, including the time array
        self.stamp = stamp
        if isinstance(self.stamp.time, Time):
            self.time = self.stamp.time
        else:
            self.time = Time(self.stamp.time, format=guess_time_format(
                self.stamp.time), scale=timescale)

    def __getitem__(self, timestep):
        '''
        Return the image data for a given timestep.

        This function is called when you say `sequence[timestep]`.

        Parameters
        ----------
        timestep : int
                A timestep index (which element in the sequence do you want?)
        '''
        if timestep is None:
            return None
        else:
            return self.stamp.todisplay[timestep, :, :]

    def _gather_3d(self):
        '''
        Gather a 3D cube of images.
        '''
        return self.stamp.todisplay[:, :, :]

    @property
    def titlefordisplay(self):
        return self.stamp.titlefordisplay

    @property
    def colorbarlabelfordisplay(self):
        return self.stamp.colorbarlabelfordisplay



from .visualize import *

def animate_temporal_differences(tpf, maxtimespan=0.02*u.day, cadence=None, sharecolorbar=False, **kw):
    '''
    For a given TPF, animate its temporal differences.
    '''

    # createa a TPF of temporal differences
    diff = tpf.differences()

    # create an illustration the original and the difference
    i = StampsIllustration([tpf, diff], sharecolorbar=sharecolorbar)
    i.plot()

    if cadence is None:
        cadence = tpf.cadence

    # animate the
    i.animate(maxtimespan=maxtimespan, cadence=cadence, **kw)

def animate_both_cadences(tpfs, filename='both_cadences.mp4', maxtimespan=30*u.minute, **kw):
    '''
    For a dictionary of TPFS, animate both the raw and the stacked images.
    '''

    raw = tpfs['raw']
    crm = tpfs['crm']
    nocrm = tpfs['nocrm']

    i = StampsIllustration([raw, crm, nocrm], names=['raw', 'CRM', 'no CRM'], sharecolorbar=False)
    for k, f in i.frames.items():
        f.titlefordisplay = '{}\n{}'.format(k, f.titlefordisplay)
    i.plot()

    i.animate(cadence=raw.cadence, maxtimespan=maxtimespan, filename=filename, **kw)


def animate_cosmics(tpfs, filename='mitigated_cosmics.mp4', maxtimespan=0.25*u.day, **kw):
    '''
    For a dictionary of TPFs, animate the mitigated cosmic rays.
    '''

    crm = tpfs['crm']
    nocrm = tpfs['nocrm']

    # make a difference TPF
    diff = copy.deepcopy(crm)
    diff.hdu[1].data['FLUX'] = nocrm.raw_cnts - crm.raw_cnts
    i = StampsIllustration([crm, nocrm, diff], names=['CRM', 'no CRM', 'difference'], sharecolorbar=False)
    for k, f in i.frames.items():
        f.titlefordisplay = '{}\n{}'.format(k, f.titlefordisplay)
    i.plot()
    i.animate(cadence=crm.cadence, maxtimespan=maxtimespan, filename=filename, **kw)



from ...imports import *
from ...postage import *
from .processing import *


def create_tpfs(s, stacker=Central(10), cadences=[2, 120,1800]):
    '''
    Starting from a Stamp, create stacked tpfs at multiple cadences.
    '''


    tpfs = {c:{} for c in cadences + [2]}
    tpfs[2]['raw'] = EarlyTessTargetPixelFile.from_stamp(s)
    for cadence in cadences:
        try:
            tpfs[cadence]['crm'] = EarlyTessTargetPixelFile.from_stamp(s.stack(cadence, strategy=stacker))
        except ValueError:
            pass
        tpfs[cadence]['nocrm'] = EarlyTessTargetPixelFile.from_stamp(s.stack(cadence, strategy=Sum()))
    return tpfs


def ok_jitter(lc, threshold=0.15, visualize=False):
    '''
    Create a mask of where the jitter is OK-ish.
    '''
    def off(x):
        return (x-np.median(x))

    ok = np.sqrt(off(lc.centroid_col)**2 + off(lc.centroid_row)**2) < threshold
    if visualize:
        fi, ax = plt.subplots(2, 1)
        for row, y in enumerate([lc.centroid_col, lc.centroid_row]):
            for which in [ok, ~ok]:
                ax[row].scatter(lc[which].time, y[which])
        ax[0].set_ylabel('col (pix)'); ax[1].set_ylabel('row (pix)')

    return ok

def save(tpfs, lcs, summary, jitter, directory):

    # save the summary
    np.save(os.path.join(directory, 'summary.npy'), summary)

    # save the jitter
    np.save(os.path.join(directory, 'jitter.npy'), jitter)

    # save the TPFs
    #if subdirectory == 'entire':
    for k in tpfs.keys():
        if k != 'raw':
            tpfs[k].to_fits(output_fn=os.path.join(directory, '{}_{}.fits'.format(tpfs[k].filelabel(), k)), zip=True)

    # save the light curves
    for k in ['crm', 'nocrm']:
        lc = lcs['{}-original'.format(k)]
        filename = os.path.join(directory, tpfs['crm'].filelabel() + '_' + k + '.csv')
        with open(filename, 'w') as f:
            f.write(lc.to_csv())

def make_tpfs(tpf2s, strategy=Central(10), cadence=1800):

    # create a stamp, for stacking into new TPFs
    s = Stamp(tpf2s)
    crm = EarlyTessTargetPixelFile.from_stamp(s.stack(cadence, strategy=strategy))
    nocrm = EarlyTessTargetPixelFile.from_stamp(s.stack(cadence, strategy=Sum()))

    return crm, nocrm


def evaluate_strategy(tpf2s,
                      directory='.',
                      strategy=Central(10),
                      cadence=1800,
                      start=-np.inf, end=np.inf,
                      aperturekw={},
                      flattenkw={},
                      correctkw={},
                      flattentimescale=0.25,
                      ):
    '''
    Evaluate a cosmic ray mitigation strategy on a TPF.

    Parameters
    ----------

    tpf2s : a TargetPixelFile object
        This is a TPF, ideally of a fairly isolated star.
        It should be at 2s cadence.

    strategy : a Stacker object
        The strategy to be used for mitigating cosmic rays.
        It defaults to a Central 8/10 strategy.

    cadence : int
        The cadence on which to evaluate the performance.

    start : float
        Minimum time to include (JD)

    end : float
        Maximum time to include (JD)

    flattentimescale : float
        The window size for the SG filter in flatten,
        expressed in days. This will be converted to
        the closest odd integer number of cadences.

    flattenkw : dict
        keyword arguments for .flatten()

    correctkw : dict
        keyword arguments for .correct()

    aperturekw : dict
        keyword arguments for define_apertures

    flattenkw : dict

    Returns
    -------

    tpfs : dict
        'crm' and 'nocrm' TPFs
    lcs : dict
        'crm' and 'nocrm' light curves with various processing
    summary : dict
        Summary of noise properties.
    jitter : dict
        a jitter timeseries
    '''
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore')

        # make the TPFs
        tpfs = {}
        tpfs['raw'] = tpf2s

        # create a stamp, for stacking into new TPFs
        s = Stamp(tpf2s)
        tpfs['crm'], tpfs['nocrm'] = make_tpfs(tpf2s, strategy=strategy)

        # define the apertures for photometry and background subtraction
        aperture, backgroundaperture = define_apertures(tpfs['nocrm'], **aperturekw)
        for k in tpfs.keys():

            # use the same aperture for all of them
            change_pipeline_aperture(tpfs[k], aperture, backgroundaperture)

            # subtract the background, redefining "flux" in each of these tpfs
            subtract_background(tpfs[k])



        # create light curve from the raw 2-s cadence
        raw = tpfs['raw'].to_lightcurve()

        # keep track of some summary statistics
        summary = {}
        summary['medflux'] = np.median(raw.flux)/raw.cadence.to('s').value

        ########################
        # KLUDGE (for bright stars)
        if (summary['medflux'] > 5e4) and 'apertureradius' not in aperturekw:
            # define the apertures for photometry and background subtraction
            aperturekw['apertureradius'] = 5 + np.sqrt(summary['medflux']/5e4)
            aperture, backgroundaperture = define_apertures(tpfs['nocrm'], **aperturekw)
            for k in tpfs.keys():

                # use the same aperture for all of them
                change_pipeline_aperture(tpfs[k], aperture, backgroundaperture)

                # subtract the background, redefining "flux" in each of these tpfs
                subtract_background(tpfs[k])

            # create light curve from the raw 2-s cadence
            raw = tpfs['raw'].to_lightcurve()

            # keep track of some summary statistics
            summary = {}
            summary['medflux'] = np.median(raw.flux)/raw.cadence.to('s').value

        lcs = {}
        lcs['raw'] = raw.normalize()

        # figure out an appropriate flatten window_length
        flattenkw['window_length'] = int(np.round(flattentimescale*24.0*3600.0/cadence))
        if (flattenkw['window_length'] % 2) == 0:
            flattenkw['window_length'] += 1
        correctkw['restore_trend'] = True

        # create all kinds of lightcurves
        for k in ['crm', 'nocrm']:

            # make a basic light curve
            lcs['{}-original'.format(k)] = tpfs[k].to_lightcurve().normalize()

            # flatten the light curve by removing temporal trends
            try:
                lcs['{}-flattened'.format(k)] = lcs['{}-original'.format(k)].remove_outliers().flatten(**flattenkw)
            except:
                print("{}-flattened lightcurve couldn't be made".format(k))
                lcs['{}-flattened'.format(k)] = lcs['{}-original'.format(k)].remove_outliers()

            # attempt to correct the light curve using SFF
            try:
                lcs['{}-corrected'.format(k)] = lcs['{}-original'.format(k)].remove_outliers().correct(**correctkw).flatten(**flattenkw)
            except:
                print("{}-corrected lightcurve couldn't be made".format(k))
                lcs['{}-corrected'.format(k)] = lcs['{}-flattened'.format(k)]

        # trim this timespan
        for k in lcs.keys():
            lc = lcs[k]

            # trim to time
            goodtime = (lc.time >= start)*(lc.time<=end)
            lcs[k] = lcs[k][goodtime]

        # record the jitter
        jitter = {}
        jt, jc, jr, jce, jre = bin_jitter(lcs['raw'], binwidth=(cadence*u.s/u.day).decompose(), robust=False)
        jitter['time'] = jt
        jitter['column'] = jc
        jitter['row'] = jr
        jitter['intraexposure'] = np.sqrt(jce**2 + jre**2)


        summary['name'] = strategy.name
        summary['ntotal'] = len(lcs['crm-original'].flux)
        for k in lcs.keys():
            if k == 'raw':
                continue
            lc = lcs[k]


            # mark the bad jitter moments
            badjitter = ok_jitter(lc) == False
            lc.quality = lc.quality | (4*badjitter)

            #    goodjitter = np.ones_like(lc.flux).astype(np.bool)
            #    lcs[k] = lcs[k][goodjitter]

            # calculate the RMS of each light curve
            ok = lc.quality == 0
            summary['{}-madstd-{:.0f}m'.format(k, cadence/60)] = mad_std(lc.flux[ok])
            summary['{}-std-{:.0f}m'.format(k, cadence/60)] = np.std(lc.flux[ok])
            clipped = sigma_clip(lc.flux[ok], sigma=3)
            fnotclipped = 1- np.sum(clipped.mask)/summary['ntotal']
            summary['{}-fractionnotclipped-{:.0f}m'.format(k, cadence/60)] = fnotclipped
            summary['{}-clippedstd-{:.0f}m'.format(k, cadence/60)] = np.std(clipped)/np.sqrt(fnotclipped)

            # calculate the equivalent for 30 minutes
            if cadence == 120:
                bx, by, bz = binto(lc.time[ok], lc.flux[ok], binwidth=0.5/24.0, robust=False, sem=True)
                summary['{}-madstd-30m'.format(k)] = mad_std(by)
                summary['{}-std-30m'.format(k)] = np.std(by)
                clipped = sigma_clip(by, sigma=3)
                fnotclipped = 1 - np.sum(clipped.mask)/summary['ntotal']
                summary['{}-fractionnotclipped-30m'.format(k)] = fnotclipped
                summary['{}-clippedstd-30m'.format(k)] = np.std(clipped)/np.sqrt(fnotclipped)

            # calculate mad-std of binned differences
            for hours in [1, 3, 6, 9, 12]:
                bx, by, bz = binto(lc.time[ok], lc.flux[ok], binwidth=hours/24.0, robust=True)
                summary['{}-d{}hr-madstd'.format(k, hours)] = mad_std(np.diff(by))
                summary['{}-d{}hr-std'.format(k, hours)] = np.std(np.diff(by))
                clipped = sigma_clip(np.diff(by), sigma=3)
                fnotclipped = 1 - np.sum(clipped.mask)/summary['ntotal']
                summary['{}-d{}hr-clippedstd'.format(k, hours)] = np.std(clipped)/np.sqrt(fnotclipped)

        if np.isfinite(start):
            summary['start'] = start
        else:
            summary['start'] = np.min(lcs['raw'].time)

        if np.isfinite(end):
            summary['end'] = end
        else:
            summary['end'] = np.max(lcs['raw'].time)
        summary['title'] = tpfs['crm'].filelabel().replace('_', ' | ') + ' | ' + strategy.name


        if np.isfinite(start) or np.isfinite(end):
            summary['title'] += ' | {:.5f} to {:.5f}'.format(start, end)

        # make sure the directories exist
        summary['directory'] = os.path.join(directory, strategy.name.replace(' ', ''))
        mkdir(summary['directory'])
        summary['directory'] = os.path.join(summary['directory'], tpfs['crm'].filelabel())
        mkdir(summary['directory'])
        if np.isfinite(start) or np.isfinite(end):
            subdirectory = '{:.5f}to{:.5f}'.format(start, end)
        else:
            subdirectory = 'entire'
        summary['directory'] = os.path.join(summary['directory'], subdirectory)
        mkdir(summary['directory'])

        # save everything
        save(tpfs, lcs, summary, jitter, directory=summary['directory'])

        return tpfs, lcs, summary, jitter



from ..imports import *
from ..sequences import *
from ..frames import *
from ..colors import cmap_norm_ticks
from ..utilities import *

class IllustrationBase(Talker):
    '''
    This contains the basic layout and organization
    for a linked visualization of images.

    Other illustrations inherit from this basic one,
    changing the
    '''
    illustrationtype = 'Base'


    def __init__(self, nrows=1,
                       ncols=1,
                       figkw=dict(figsize=None, dpi=None),
                       sharecolorbar=True,
                       subplot_spec=None,
                       cmapkw=dict(),
                       **gridspeckw):
        '''
        Initialize an Illustration,
        setting up its figure and basic layout.

        nrows : int
            The number of rows in the illustration.

        ncols : int
            The number of columns in the illustration.

        figkw : dict
            Keywords to pass along to create a figure.

        **gridspeckw : dict
            Any extra keywords are passed to the
            gridspec
        '''

        Talker.__init__(self, prefixformat='{:>32}')

        if subplot_spec is not None:
            # if there's a subplot_spec specified, then populate that
            self.figure = None

            subset = {k:gridspeckw.get(k) for k in ['wspace', 'hspace', 'height_ratios', 'width_ratios']}
            self.grid = gs.GridSpecFromSubplotSpec(nrows,
                                                    ncols,
                                                    subplot_spec=subplot_spec,
                                                    **subset)
            #self.speak('built {} into an existing gridspec'.format(self.illustrationtype))
        else:
            # by default, create a new figure and grid spec
            self.figure = plt.figure(**figkw)
            self.grid = gs.GridSpec(nrows, ncols, **gridspeckw)
            #self.speak('built {} into a new figure'.format(self.illustrationtype))


        # should this illustration have a shared colorbar, or separate ones?
        self.sharecolorbar = sharecolorbar

        # keep track of any keywords for generating a cmap (shared for the illustration)
        self.cmapkw = copy.copy(cmapkw) # why do I have to do this?

        # create an empty dictionary, where frames will be stored
        self.frames = {}

        # has this illustration been plotted yet?
        self.hasbeenplotted = False

    def __repr__(self):
        '''
        How should this illustration be represented?
        '''
        return '<{} Illustration | ({} Frames)>'.format(self.illustrationtype, len(self.frames))

    def __str__(self):
        return '<I"{}">'.format(self.illustrationtype)

    def _get_times(self):
        '''
        Get *all* the times that are associated with this
        illustration, in any frame.
        '''

        # it's faster to convert from gps to floats and back again

        # make a list of gps times
        gps = [f._get_times().gps for f in self.frames.values()]

        # return that list, as times
        return Time(np.hstack(gps), format='gps', scale='tdb')

    def _timesandcadence(self, round=1):
        '''
        Get all the unique times available across all the frames,
        along with a suggested cadence set by the minimum
        (rounded) differences between times.

        Parameters
        ----------
        round : float
                All times will be rounded to this value.
                Times separated by less than this value
                will be considered identical.

                It is in units of seconds.
        '''

        try:
            self._precaculatedtimesandcadence
        except AttributeError:
            self._precaculatedtimesandcadence = {}
        try:
            times, cadence = self._precaculatedtimesandcadence[round]
        except KeyError:

            gps = np.hstack([f._get_times().gps for f in self.frames.values()])
            # alltimes =  Time(np.hstack(gps), format='gps')
            # alltimes = []
            # for k, f in self.frames.items():
            #	alltimes.extend(f._get_times())

            if round is None:
                diffs = np.diff(np.sort(gps))
                round = np.min(diffs[diffs > 0])

            baseline = np.min(gps)
            rounded = round * np.round((gps - baseline) / round) + baseline
            uniquegpstimes = np.unique(rounded)

            # make sure there's a cadence, even if there's only one element in the sequence
            try:
                cadence = np.min(np.diff(uniquegpstimes)) * u.s
            except ValueError:
                cadence = 1.0*u.s

            # plt.figure('cadence!')
            # plt.plot(uniquegpstimes[:-1], np.diff(uniquegpstimes), '.')
            # plt.show()
            times = Time(uniquegpstimes, format='gps')
            self._precaculatedtimesandcadence[round] = times, cadence
        return times, cadence

    def plot(self, *args, **kwargs):
        '''
        Plot all the frames in this illustration.

        *args, **kwargs are passed to frames' .plot()
        '''

        self.plotted = {}
        for k, f in self.frames.items():
            f.plot(*args, **kwargs)
        self.hasbeenplotted = True


    def update(self, *args, **kwargs):
        '''
        Update all the frames in this illustration.

        *args, **kwargs are passed to the frames' .update()
        '''
        for k, f in self.frames.items():
            f.update(*args, **kwargs)

    def _cmap_norm_ticks(self, remake=False, **cmapkw):
        '''
        Return the cmap and normalization.

        Make a colorbar for this illustration,
        using the first image from every frame.

        **kwargs are passed to colors.cmap_norm_ticks
        '''

        try:
            # are the cmap, normalization, and ticks already defined?
            (self.plotted['cmap'],
             self.plotted['norm'],
             self.plotted['ticks'])

            assert(remake == False)
        except (KeyError, AssertionError):

            # collect all the (first) images
            firstimages = []
            for name, frame in self.frames.items():
                try:
                    firstimages.extend(frame._get_image()[0].flatten())
                    self.speak('included {} in the shared color scheme'.format(frame))
                except (TypeError, IndexError, AttributeError):
                    self.speak('found no color scheme data for {}'.format(frame))

            # create the cmap from the given data
            (self.plotted['cmap'],
             self.plotted['norm'],
             self.plotted['ticks']) = cmap_norm_ticks(np.asarray(firstimages),
                                                      **cmapkw)
            self.speak('defined color scheme with \n cmap={}\n norm={}\n ticks={}'.format(self.plotted['cmap'],
                                                                                          self.plotted['norm'],
                                                                                          self.plotted['ticks']))
        return  (self.plotted['cmap'],
                 self.plotted['norm'],
                 self.plotted['ticks'])

    def _add_colorbar(self, imshowed, ax=None, ticks=None):
        '''
        Make a colorbar,
        attached to a particular axes
        (or multiple axes at once).

        Parameters
        ----------
        imshowed : the output of plt.imshow
                the colorbar will be extracted from this imshow
        ax : axes object
                this colorbar will be attached to this axes

        '''

        # (this means we have a shared colorbar)
        if ax is None:
            # pull all the different ax together
            ax = [f.ax for f in self.frames.values() if isinstance(f, imshowFrame)]
            # if an imshow doesn't have its own actual ax, don't include it
            ax = [a for a in ax if a != None]

        self.speak('adding a new colorbar for {} frame(s)'.format(len(np.atleast_1d(ax))))

        # make a colorbar attached to the frame
        #	axes =
        colorbar = plt.colorbar(
            imshowed,
            ax=ax,
            orientation='horizontal',
            # label=self.data.colorbarlabelfordisplay,
            fraction=0.04,
            pad=0.08,
            ticks=ticks)

        # colorbarred = plt.colorbar(self.plotted['image'], ax=axes, orientation='horizontal', label=self.data.colorbarlabelfordisplay, fraction=0.04, pad=0.07, ticks=ticks)
        colorbar.ax.set_xticklabels(
            ['{:.0f}'.format(v) for v in ticks], fontsize=8, color='gray')
        colorbar.outline.set_visible(False)

        return colorbar

    def _condense_timelabels(self):
        '''
        Remove all but a single time label.
        Don't do this if you have staggered time-series.
        '''

        # remove all but the lower left time indicator
        x, y = np.inf, np.inf
        N = 0
        for k in self.frames.keys():
            if 'time' in self.frames[k].plotingredients:
                # remove the time label
                self.frames[k].plotingredients.remove('time')
                N +=1

                # keep track of whether this is the lower left corner
                bbox = self.frames[k].ax.get_position()
                if bbox.x0 <= x and bbox.y0 <=y:
                    corner = k
                    x, y = bbox.x0, bbox.y0

        # add a time to the lower left corner
        self.frames[corner].plotingredients.append('time')
        self.speak('condensing {} time labels into 1'.format(N))

    def savefig(self, filename, **savefigkw):
        '''
        Call plt.savefig, but make an announcement about it.
        This should be called after illustration.plot()

        Parameters
        ----------

        filename : str
            Where should the figure be saved?

        **savefigkw are passed to savefig
        '''

        self.speak('saving figure to {}'.format(filename))
        plt.savefig(filename, **savefigkw)
        self.speak('saved figure to {}'.format(filename))


    def animate(self, filename='test.mp4',
                      mintime=None, maxtimespan=None, cadence=1 * u.s,
                      fps=30, dpi=None, **kw):
        '''
        Create an animation from an Illustration,
        using the time axes associated with each frame.

        This should be called after illustration.plot()

        Parameters
        ----------

        filename : str
        '''

        if self.hasbeenplotted == False:
            self.plot()

        # figure out the times to display
        if mintime is None:
            actualtimes, actualcadence = self._timesandcadence(
                round=cadence.to('s').value)
            lower, upper = min(actualtimes.gps), max(
                actualtimes.gps) + actualcadence.to('s').value
        else:
            lower = mintime.gps
            upper = lower + maxtimespan.to('s').value

        if cadence is None:
            cadence = actualcadence.to('s').value
        else:
            # np.maximum(cadence.to('s').value, actualcadence.to('s').value)
            cadence = cadence.to('s').value

        if maxtimespan is not None:
            upper = lower + np.minimum(upper - lower, maxtimespan.to('s').value)

        times = np.arange(lower, upper, cadence)
        self.speak('about to animate {} times at {}s cadence for {}'.format(
            len(times), cadence, self))

        # get the writer
        writer = get_writer(filename, fps=fps)

        self.speak('the animation will be saved to {}'.format(filename))
        # set up the animation writer
        with writer.saving(self.figure,
                           filename,
                           dpi or self.figure.get_dpi()):

            for i, t in enumerate(times):
                self.speak('  {}/{} at {}'.format(i + 1,
                            len(times), Time.now().iso), progress=True)

                # update the illustration to a new time
                self.update(Time(t, format='gps', scale='tdb'))
                writer.grab_frame()
        self.speak('')
        self.speak('the animation is finished!')

"""
class Row(IllustrationBase):
	'''
	FIXME -- the initialization is probably still sketchy on this.
	'''

	def __init__(self, *frameargs, **framekwargs):

		# store all the frames in a dictionary
		self.frames = {}
		for i, a in enumerate(frameargs):
			self.frames[i] = a

		for i, k in enumerate(framekwargs.keys()):
			self.frames[k] = framekwargs[k]

		# set up the geometry of the figure
		self.orientation = 'horizontal'
		cols = len(self.frames)
		rows = 1
		IllustrationBase.__init__(self, rows, cols,
								hspace=0.02, wspace=0.02,
								left=0.05, right=0.95,
								bottom=0.1, top=0.9)

		# initiate the axes for each camera
		keys = list(range(len(frameargs))) + list(framekwargs.keys())
		for j, k in enumerate(keys):

			# populate the axes on the main camera grid
			ax = plt.subplot(self.grid[0, j])
			self.frames[k].ax = ax
"""



from .IllustrationBase import *

__all__ = ['GenericIllustration']


class SideBySideIllustration(IllustrationBase):
    illustrationtype = 'Timeseries'

    def __init__(self, imshows=[], timeseries=[], width_ratios = [5, 1], each = 1.0,
                                hspace=0.15, wspace =  0.08, left=0.15, right =  0.95,
                                bottom = 0.3, top=0.9, fudge=1.2, **kwargs):
        '''
        Initialize an illustration from list of frames.
        '''

        assert(len(imshows) == len(timeseries))

        # set up the geometry (imshows on time, timeseries below)
        rows = len(imshows)
        cols = 2


        wsize = each * np.sum(width_ratios) * \
            (1 + (cols - 1) * wspace) / (right - left)
        hsize = fudge * each * rows * (1 + (rows - 1) * hspace) / (top - bottom)

        IllustrationBase.__init__(self, rows, cols,
                                  figkw=dict(figsize=(wsize, hsize)),
                                  hspace=hspace, wspace=wspace,
                                  left=left, right=right,
                                  bottom=bottom, top=top,
                                  width_ratios=width_ratios,
                                  **kwargs)

        # add the imshows
        for row, i in enumerate(imshows):
            col = 1
            # make sure to connect the frame back to this illustration
            i.illustration = self
            # create the axes in which this should sit
            i.ax = plt.subplot(self.grid[row, col])

            # register this in the frames dictionary
            try:
                n = i.name
            except AttributeError:
                n = (row, col)
            self.frames[n] = i

        # add the timeseries
        sharex = None  # you may want to change the shared axes for different visualizations
        for row, t in enumerate(timeseries):
            col = 0

            # make sure to connect the frame back to this illustration
            t.illustration = self
            # create the axes in which this should sit
            t.ax = plt.subplot(self.grid[row, col], sharex=sharex)
            sharex = t.ax

            # turn off the axis labels
            if row != len(timeseries) - 1:
                plt.setp(t.ax.get_xticklabels(), visible=False)
            # register this in the frames dictionary
            try:
                n = t.name
            except AttributeError:
                n = (row, col)
            self.frames[n] = t



from .IllustrationBase import *
from .CameraIllustration import *
from ..frames import CameraFrame, cameras

__all__ = ['FourCameraIllustration']


class FourCameraIllustration(IllustrationBase):
    '''
    For displaying a single Camera.
    '''
    illustrationtype = 'FourCamera'

    def __init__(self, cam1=[], cam2=[], cam3=[], cam4=[], orientation='horizontal', sizeofcamera=4, subplot_spec=None, **kwargs):
        '''

        Parameters
        ----------
        cam1, etc... : various
            Anything that can initialize a sequence with
            the make_image_sequence() helper function.

        orientation : str
            'horizontal' = camera 2 is right of camera 1
            'vertical'  = camera 2 is below camera 1 (not yet implemented)

        sizeofcamera : float
            What's the size, in inches, to display a single camera?
        '''

        # set up the basic geometry of the main axes
        sizeofcamera = 4
        N = 4
        self.orientation = orientation
        if self.orientation == 'horizontal':
            cols = N
            rows = 1
        IllustrationBase.__init__(self, rows, cols,
                                  figkw=dict(
                                      figsize=(sizeofcamera * cols, sizeofcamera * rows * 1.2)),
                                  hspace=0.02, wspace=0.02,
                                  left=0.05, right=0.95,
                                  bottom=0.1, top=0.9,
                                  subplot_spec=subplot_spec)

        # initiate the axes for each camera
        for i in range(rows):
            for j in range(cols):

                # create a CameraFrame for this camera
                n = i * cols + j + 1
                name = 'cam{}'.format(i * cols + j + 1)

                # populate the axes on the main camera grid
                ax = plt.subplot(self.grid[i, j])

                self.frames[name] = cameras[name](
                    ax=ax, data=locals()[name], illustration=self, **kwargs)



from ... import *
from ...imports import *

def plot_aperture_definition(tpf):
    '''
    For a given TPF, visualize its aperture definitions.
    '''

    aspect = tpf.shape[1]/float(tpf.shape[2])
    fi, ax = plt.subplots(1,3,figsize=(10, 1.8*aspect), sharex=True,sharey=True)
    frame = int(len(tpf.time)/2)
    tpf.plot(ax[0], frame=frame, bkg=True, clabel='')
    tpf.plot(ax[1], aperture_mask=tpf.hdu[-1].data == 2, frame=frame, bkg=True, clabel='')
    ax[1].set_title('background')
    tpf.plot(ax[2], aperture_mask=tpf.hdu[-1].data == 3, frame=frame, bkg=False, clabel='')
    ax[2].set_title('target')


def plot_histogram(y, bins=None, binwidth=0.1, ax=None, **kwargs):
    '''
    Plot a histogram, rotated clockwise by 90 degrees, to represent a projection of a timeseries plot.

    Parameters
    ----------

    y : array
        The y values to go into the histogram.

    bins : array
        The array of locations for the histogram (otherwise, requires binwidth)

    binwidth : float
        If bins aren't set, create a grid using this binwidth and the sample limits.

    ax : an Axes object
        (optional) Plot into a particular ax.

    **kwargs are passed to plt.plot()
    '''

    # create a histogram of the lightcurve values
    if bins is None:
        bins = np.arange(np.min(y)-binwidth, np.max(y)+binwidth, binwidth)

    ok = np.isfinite(y)
    yhist, edges = np.histogram(y[ok], bins=bins, density=True)

    # define the "x"-axis at the centers of the histogram bins
    xhist = (edges[1:] + edges[0:-1])/2.0
    binwidth = np.median(np.diff(xhist))

    # plot in the histogram panel
    if ax is None:
        ax = plt.gca()
    ax.plot(yhist, xhist, **kwargs)

    ax.set_xscale('log')
    ax.set_xlim(5.0/binwidth/len(y), 10*1.0/np.sqrt(2*np.pi)/binwidth)#np.max(yhist)*2)
    plt.axis('off')



def plot_timeseries(x, y, ax, ylim, ylabel='', color='black', alpha=1, **kw):
    '''
    Plot a single timeseries (with both 1D series and a collapsed histogram).
    '''
    try:
        plt.sca(ax[0])
    except IndexError:
        plt.sca(ax)
    # plot the timeseries
    plt.plot(x, y, color=color, alpha=alpha, **kw)

    # plot any outliers
    above = y > np.max(ylim)
    below = y < np.min(ylim)
    scale = 0.2*(np.max(ylim) - np.min(ylim))
    outlierkw = dict(**kw)
    outlierkw['alpha'] = 0.2
    outlierkw['elinewidth'] = 1
    outlierkw['linewidth'] = 0
    outlierkw['capsize'] = 0
    outlierkw['marker'] = None
    plt.errorbar(x[above], np.max(ylim)*np.ones_like(x[above]), scale, color=color, **outlierkw)
    plt.errorbar(x[below], np.min(ylim)*np.ones_like(x[below]), scale, color=color, **outlierkw)
    #plt.ylabel(ylabel)
    if np.isfinite(ylim).all():
        plt.ylim(*ylim)


    # plot the histogram
    try:
        plt.sca(ax[1])
        nbins = np.sqrt(len(y))
        plot_histogram(y, bins='auto', color=color, alpha=alpha) #bins=np.linspace(np.min(ylim), np.max(ylim), nbins)
        plt.ylim(*ylim)
    except IndexError:
        pass

def visualize_strategy(tpfs, lcs, summary, jitter, animation=False, nsigma=5, ylims={}, **kw):

    # create panels for the images we want to show
    i_nocrm = imshowFrame(data=tpfs['nocrm'], name='nocrm')
    i_nocrm.titlefordisplay = 'without CRM'

    i_crm = imshowFrame(data=tpfs['crm'], name='crm')
    i_crm.titlefordisplay = 'with CRM'

    diff = copy.deepcopy(tpfs['crm'])
    diff.hdu[1].data['FLUX'] = tpfs['nocrm'].raw_cnts - tpfs['crm'].raw_cnts
    i_diff = imshowFrame(data=diff, name='difference')
    i_diff.titlefordisplay = 'difference'

    i_aperture = imshowFrame(data=tpfs['crm'], name='aperture')
    i_aperture.titlefordisplay = '(aperture definition)'
    imshows = [i_nocrm, i_crm, i_diff, i_aperture]

    # create panels for timeseries to show
    timeseries = [EmptyTimeseriesFrame(name='original', ylabel='original\nflux', histogram=True),
                  EmptyTimeseriesFrame(name='flattened', ylabel='flattened\nflux', histogram=True),
                  EmptyTimeseriesFrame(name='corrected', ylabel='corrected\nflux', histogram=True),
                  EmptyTimeseriesFrame(name='fluxdiff', ylabel='CRM flux\ndifference', histogram=True),
                  EmptyTimeseriesFrame(name='column', ylabel='centroid\ncolumn\n(pix)', histogram=False),
                  EmptyTimeseriesFrame(name='row', ylabel='centroid\nrow\n(pix)', histogram=False),
                  EmptyTimeseriesFrame(name='intraexposure', ylabel='intra-\nexposure\njitter\n(pix)', histogram=False)]

    # put them all together
    i = GenericIllustration(imshows=imshows, timeseries=timeseries, sharecolorbar=False)
    i.plot()

    # add the aperture onto its imshow frame
    mask = tpfs['crm'].pipeline_mask == False
    extent = [0, mask.shape[1], 0, mask.shape[0]]
    i_aperture.ax.imshow(mask, interpolation='nearest', origin='lower', extent=extent, cmap=one2another(alphabottom=0, alphatop=0.5, top='purple'))
    i_aperture.ax.imshow(tpfs['crm'].background_mask, interpolation='nearest', origin='lower', extent=extent, cmap=one2another(alphabottom=0, alphatop=0.5, top='hotpink'))


    # plot the individual timeseries
    colors = dict(crm='mediumvioletred', nocrm='royalblue', raw='orange')
    lckw = dict(marker='.', markersize=8, linewidth=0, markeredgecolor='none')

    scale = nsigma*np.maximum(mad_std(lcs['crm-flattened'].flux),
                              mad_std(lcs['nocrm-flattened'].flux))

    for mode in ['original', 'flattened', 'corrected']:
        f = i.frames[mode]
        for c in ['nocrm', 'crm']:
            lc = lcs['{}-{}'.format(c, mode)]
            if mode in 'original':
                if mode in ylims:
                    ylim = ylims[mode]
                else:
                    bottom, top = np.nanpercentile(lc.flux, [1,99])
                    ylim = [np.minimum(bottom, 1-scale), np.maximum(top, 1+scale)]
            else:
                ylim = 1-scale, 1+scale
            plot_timeseries(lc.time - f.offset, lc.flux, [f.ax, f.ax_hist], ylim, color=colors[c], alpha=1, **lckw)

            f.ax_hist.text(0.995, 0.97 - (1.0 + 5*(c=='crm'))/7,
                              '{}: {}={:.0f}ppm, {}={:.0f}ppm, {}={:.0f}ppm'.format(
                                    c.upper(),
                                    r'$\sigma$', 1e6*summary['{}-{}-std-{:.0f}m'.format(c, mode, lc.cadence.to('min').value)],
                                    r'$\sigma_{clipped}$', 1e6*summary['{}-{}-clippedstd-{:.0f}m'.format(c, mode, lc.cadence.to('min').value)],
                                    r'$\sigma_{MAD}$', 1e6*summary['{}-{}-madstd-{:.0f}m'.format(c, mode, lc.cadence.to('min').value)]),
                              color=colors[c], ha='right', va='center', transform=f.ax.transAxes )

    # plot the centroid information
    centroid_scale = nsigma*np.maximum(mad_std(jitter['column']),mad_std(jitter['row']))
    centroid_ylim = [-centroid_scale, centroid_scale]
    for cen in ['column', 'row']:
        f = i.frames[cen]
        plot_timeseries(jitter['time']-f.offset, jitter[cen], [f.ax], centroid_ylim, color='dimgray', **lckw)

        for c in ['nocrm', 'crm']:
            lc = lcs['{}-original'.format(c)]
            if cen == 'column':
                y = lc.centroid_col
            elif cen == 'row':
                y = lc.centroid_row
            plot_timeseries(lc.time-f.offset, y-np.median(y), [f.ax], centroid_ylim, color=colors[c], alpha=0.5, **lckw)

    ylim =  [0, 2*nsigma*mad_std(jitter['intraexposure']) + np.median(jitter['intraexposure'])]
    f = i.frames['intraexposure']
    plot_timeseries(jitter['time']-f.offset, jitter['intraexposure'],  [f.ax], ylim, color='dimgray', **lckw)
    #f.ax.set_ylim(0, None)


    # calculate the net gains and losses
    lc_diff = {}
    normalization = tpfs['crm'].to_lightcurve().flux
    gains = copy.deepcopy(diff)
    gains.hdu[1].data['FLUX'] = np.maximum(gains.flux, 0)
    lc_diff['gains']  = gains.to_lightcurve().flux/normalization
    losses = copy.deepcopy(diff)
    losses.hdu[1].data['FLUX'] = np.minimum(losses.flux, 0)
    lc_diff['losses'] = losses.to_lightcurve().flux/normalization

    #lc_diff['net'] = lc_diff['gains'] + lc_diff['losses']
    colors_diff = dict(gains='royalblue', losses='firebrick', net='dimgray')
    f = i.frames['fluxdiff']
    t = gains.time - f.offset
    scale = 2*nsigma*mad_std(lc_diff['gains'] + lc_diff['losses'])
    ylim = [-scale, scale]
    for k, l in lc_diff.items():
        plot_timeseries(t, l,
                        ax=[f.ax, f.ax_hist],
                        ylim=ylim,
                        color=colors_diff[k],
                        linewidth=1, alpha=0.5)
    plot_timeseries(t, lc_diff['gains'] + lc_diff['losses'],
                    ax=[f.ax, f.ax_hist],
                    ylim=ylim,
                    color='dimgray',
                    **lckw)
    for a in [f.ax, f.ax_hist]:
        a.axhline(0, color='gray', alpha=0.3)

    title = ' | '.join([imshows[0].data.titlefordisplay.replace('\n', ' | '), summary['name'], '{:.2g} DN/s'.format(summary['medflux'])])
    plt.suptitle(title, fontsize=20)

    for t in timeseries:
        try:
            t.ax.set_xlim(summary['start'] - t.offset, summary['end'] - t.offset)
        except (ValueError, TypeError):
            pass

    d = summary['directory']
    filename = 'tic{}_{:.0f}m_{}'.format(tpfs['crm'].tic_id, tpfs['crm'].cadence.to('minute').value, summary['name'].replace(' ',''))
    plt.savefig(os.path.join(d, filename+'.pdf'))
    if animation:
        i.animate(cadence=tpfs['crm'].cadence, filename=os.path.join(d, filename+'.mp4'), mintime=Time(summary['start'], format='jd', scale='tdb'), maxtimespan=(summary['end'] - summary['start'])*u.day, **kw)
    return i



def plot_jitter_correlate(path):
    '''
    Make and save a jitter correlation plot for a given directory path
    (that must contain some light curves, a jitter, and a summary).
    '''
    label = path.split('/')[-3].replace('_', ' | ')
    lcfiles = glob.glob(os.path.join(path, '*.csv'))
    lcs = {}
    for f in lcfiles:
        k = f.split('_')[-1].split('.csv')[0]
        lcs[k] = ascii.read(f, delimiter=',', data_start=1, names='time,time1,flux,flux_err,quality,centroid_col,centroid_row'.split(','))
    diff = lcs['nocrm']['flux'] - lcs['crm']['flux']

    jitterfile = os.path.join(path, 'jitter.npy')
    jitter = np.load(jitterfile, encoding='latin1')[()]


    summaryfile = os.path.join(path, 'summary.npy')
    summary = np.load(summaryfile, encoding='latin1')[()]

    rows = dict(crm=lcs['crm']['flux'], nocrm=lcs['nocrm']['flux'])
    rows['nocrm-crm'] = diff
    # make a plot
    fi, ax = plt.subplots(3, 4, sharey='row', sharex='col', figsize=(10, 7), dpi=200,
                         gridspec_kw=dict(hspace=0.04, wspace=0.04))
    for row, ylabel in enumerate(['nocrm', 'crm', 'nocrm-crm']):
        for col, k in enumerate(['row', 'column', 'intraexposure', 'time']):
            plt.sca(ax[row, col])
            x = jitter[k][:len(diff)]
            if k == 'time':
                x-= np.min(x)
            y = rows[ylabel]

            xlim = np.nanpercentile(x[np.isfinite(x)], [2, 98])
            ylim = np.nanpercentile(y[np.isfinite(y)], [2, 98])
            plt.plot(x, y, '.', alpha=0.5, markeredgecolor='none', color='mediumseagreen')

            ok = (x >= xlim[0])*(x <= xlim[1])*(y >= ylim[0])*(y <= ylim[1])
            nbins=20
            bx, by, be = binto(x[ok], y[ok], binwidth=(xlim[1] - xlim[0])/nbins, robust=True)
            plt.errorbar(bx, by, be, color='black', linewidth=0, elinewidth=4, zorder=10)


            plt.axhline(0, color='black', zorder=100, alpha=0.5)
            if col == 0:
                plt.ylabel(ylabel)#('Fractional flux\nlost to CRM')
            if row == 2:
                plt.xlabel(k)

            rho = scipy.stats.spearmanr(x[ok],y[ok])[0]
            R = scipy.stats.pearsonr(x[ok],y[ok])[0]
            plt.text(0.05, 0.95, 'R={:.3f}\n{}={:.3f}'.format(R, r'$\rho$', rho), transform=plt.gca().transAxes, ha='left', va='top')
            if k != 'intraexposure' or ylabel != 'nocrm-crm':
                plt.gca().set_facecolor('gainsboro')

            plt.xlim(*xlim)
            plt.ylim(*ylim)
    plt.suptitle(summary['title'].replace('s |', 's\n') + ' | {:.2g} DN/s'.format(summary['medflux']))
    #plt.tight_layout()
    filename = 'jitter_' + summary['title'].replace(' | ', '_').replace(' ', '') + '.pdf'
    plt.savefig(os.path.join(path, filename))
    print(os.path.join(path, filename))



'''
Make some cartoon datasets to play with.
'''

from .imports import *
from .postage.stamps import Stamp
from .postage.tpf import EarlyTessTargetPixelFile, EarlyTessLightCurve

def create_test_times(N=100, cadence=2):
    '''
    Create a fake array of times.
    '''
    return Time('2018-01-01 00:00:00.000', scale='tdb') + np.arange(N) * cadence * u.s


def create_test_array(N=100, xsize=5, ysize=5, nstars=None, single=False, seed=None):
    '''
    Create a fake stack of images of stars
    (with no cosmics injected, no jitter, Gaussian PDF).

    Parameters:
    -----------
    xsize : int
            The number of columns.
    ysize : int
            The number of rows.
    n : int
            The number of time points.
    single : bool
            Should this be just a single star, or should it be many?
    '''

    # seed the random number generator
    np.random.seed(seed)

    # create images of x, y, and an empty one to fill with stars
    x1d = np.arange(0, xsize)
    y1d = np.arange(0, ysize)
    x, y = np.meshgrid(x1d, y1d)
    stars = np.zeros_like(x)

    # create N random position for stars
    if single:
        nstars = 1
        sx = np.random.normal(xsize / 2.0, 1)
        sy = np.random.normal(ysize / 2.0, 1)
    else:
        if nstars is None:
            nstars = np.minimum(int(xsize * ysize / 4), 1000)
        sx = np.random.uniform(0, xsize, nstars)
        sy = np.random.uniform(0, ysize, nstars)

    # set some background level
    bg = 30

    # create a semi-reasonable magnitude distribution for stars
    if single:
        topmag = 5
    else:
        topmag = 10
    sf = 10000 * 10**(-0.4 * np.random.triangular(0, topmag, topmag, nstars))

    # define the cartoon PSF for the stars
    sigma = 1.0

    def gauss(x, y, x0, y0):
        return np.exp(-0.5 * ((x - x0)**2 + (y - y0)**2) / sigma**2)

    # create a cube with an image for each star, and sum them together into one image
    stars = (sf * gauss(x[:, :, np.newaxis],
                        y[:, :, np.newaxis], x0=sx, y0=sy)).sum(2)

    # create a model images
    model = (bg + stars)[np.newaxis, :, :] * np.ones(N).reshape((N, 1, 1))

    # add some noise to it
    image = np.random.normal(model, np.sqrt(model))

    return image


def create_test_stamp(col_cent=3900, row_cent=913, cadence=2, cam=1, spm=1, tic_id=1234567890, **kw):
    static = {'CAM': cam,
              'COL_CENT': col_cent,
              'INT_TIME': cadence,
              'ROW_CENT': row_cent,
              'SPM': spm,
              'TIC_ID': tic_id}

    spatial = {}

    photons = create_test_array(**kw)
    N = photons.shape[0]
    temporal = {}
    temporal['TIME'] = create_test_times(N=N, cadence=cadence)
    temporal['CADENCE'] = np.arange(N)
    temporal['QUAL_BIT'] = np.zeros(N).astype(np.int)

    return Stamp(spatial=spatial, photons=photons, temporal=temporal, static=static)


def create_test_lightcurve(N=100, cadence=2, **kwargs):
    '''
    Create a test lightkurve LightCurve object.
    '''
    time = create_test_times(N=N, cadence=cadence).jd
    flux = np.random.normal(1, 0.001, len(time))
    return EarlyTessLightCurve(time, flux)


def create_test_tpf(**kwargs):
    '''
    Create a test TargetPixelFile.

    Parameters
    ----------
    '''
    return EarlyTessTargetPixelFile.from_stamp(create_test_stamp(**kwargs))


def create_test_fits(rows=400, cols=600, circlescale=100, visualize=False, noise=0.2, seed=None):
    '''
    Create a test FITS image, mostly to help understand geometry.
    It has circles centered at 0 and a gradient toward positive x.
    '''

    # create x and y arrays
    x, y = np.meshgrid(np.arange(cols).astype(np.float),
                       np.arange(rows).astype(np.float))

    # put a dot near the origin, and a gradient in x
    z = 0.5 * np.max(x) * np.exp(-0.5 *
                                 np.sin((x**2 + y**2) / circlescale**2)**2) + x
    if noise is not None:
        np.random.seed(seed)
        sigma = noise * np.std(z)
        z = np.random.normal(z, sigma)

    hdulist = fits.HDUList([fits.PrimaryHDU(),
                            fits.ImageHDU(z, name='image'),
                            fits.ImageHDU(x, name='x'),
                            fits.ImageHDU(y, name='y')])
    #hdulist[0].header['TIME'] = 42.0

    if visualize:
        plt.subplot(131)
        plt.imshow(x, origin='lower')
        plt.title('x')
        plt.subplot(132)
        plt.imshow(y, origin='lower')
        plt.title('y')
        plt.subplot(133)
        plt.title('z')
        plt.imshow(z, origin='lower')

    return hdulist

def create_directory_of_fits(base='someplace', N=10, xsize=100, ysize=100, **kw):

    mkdir(base)
    for camera in [1,2,3,4]:
        directory = os.path.join(base, 'cam{}'.format(camera))
        mkdir(directory)
        for i in range(N):
            filename = os.path.join(directory, 'cam{}-{:04}.fits'.format(camera, i))
            if os.path.exists(filename):
                continue
            f = create_test_fits(rows=ysize, cols=xsize, circlescale=(i+1)*6, **kw)
            f.writeto(filename, overwrite=True)
            print(' saved cartoon image to {}'.format(filename))



from .imshowFrame import *
from matplotlib.colors import SymLogNorm, LogNorm

class CameraFrame(imshowFrame):
    frametype = 'camera'

    def __init__(self, name='camera',
                       ax=None,
                       data=None,
                       **framekw):

        '''
        Initialize a CameraFrame. For plotting, this
        expects to see a full-frame image from a full
        camera (not split into CCDs.)

        Parameters
        ----------

        name : str
            A name to give this Frame.

        ax : matplotlib.axes.Axes instance
            All plotting will happen inside this ax.
            If set to None, the `self.ax attribute` will
            need to be set manually before plotting.

        data : Image_Sequence
            Any Sequence that contains 2D images.

        **framekw : dict
            Passed to imshowFrame;
            check it out for more options.
        '''

        # initialize the imshow frame
        imshowFrame.__init__(self, name=name, ax=ax, data=data, **framekw)

        # set up the basic geometry
        self.xmin, self.ymin = 0, 0
        try:
            # if there's an image, use it to set the size
            self.ymax, self.xmax = self.data[0].shape
        except (IndexError, AttributeError, TypeError):
            # allow us to create an empty image
            self.ymax = 4156
            self.xmax = 4272

        # by default, this camera frame doesn't do any transformation
        self.transpose = False
        self.flipy     = False
        self.flipx     = False

    def _transformimage(self, image):
        '''
        horizontal:
                (should be) +x is up, +y is left
                (looks like) +x is up, +y is right
        '''
        if self._get_orientation() == 'horizontal':
            if self.transpose:
                displayimage = image.T
            else:
                displayimage = image
            if self.flipy:
                displayimage = displayimage[::-1, :]
            if self.flipx:
                displayimage = displayimage[:, ::-1]
        else:
            raise RuntimeError("Sorry! No orientations besides 'horizontal' have been defined yet!")

        return displayimage

    def _transformxy(self, x, y):
        '''
        This handles the same transformation as that which goes into
        transform image, but for x and y arrays.
        '''
        if self._get_orientation() == 'horizontal':
            if self.transpose:

                if self.flipx:
                    displayx = self.ymax - y
                else:
                    displayx = y

                if self.flipy:
                    displayy = self.xmax - x
                else:
                    displayy = x

            else:
                if self.flipx:
                    displayx = self.xmax - x
                else:
                    displayx = x

                if self.flipy:
                    displayy = self.ymax - y
                else:
                    displayy = y
        else:
            raise RuntimeError("Sorry! No orientations besides 'horizontal' have been defined yet!")

        return displayx, displayy


class Camera1Frame(CameraFrame):

    def __init__(self, name='cam1', **kwargs):
        CameraFrame.__init__(self, name=name, **kwargs)
        self.transpose = True
        self.flipy     = False
        self.flipx     = False

class Camera2Frame(Camera1Frame):
    def __init__(self, name='cam2', **kwargs):
        Camera1Frame.__init__(self, name=name, **kwargs)

class Camera3Frame(CameraFrame):
    def __init__(self, name='cam3', **kwargs):
        CameraFrame.__init__(self, name=name, **kwargs)
        self.transpose = True
        self.flipy     = True
        self.flipx     = True

class Camera4Frame(Camera3Frame):
    def __init__(self, name='cam4', **kwargs):
        Camera3Frame.__init__(self, name=name, **kwargs)

# define a dictionary, so these cameras are easier to access
cameras = { 'camera': CameraFrame,
            'cam1'  : Camera1Frame,
            'cam2'  : Camera2Frame,
            'cam3'  : Camera3Frame,
            'cam4'  : Camera4Frame}



from .IllustrationBase import *
from .CameraOfCCDsIllustration import *
from ..frames import CameraFrame, cameras

__all__ = ['FourCameraOfCCDsIllustration']


class FourCameraOfCCDsIllustration(IllustrationBase):
    '''
    For displaying a single Camera, by organizing
    individual TESS CCDs into their appropriate
    places and orientations.
    '''
    illustrationtype = 'FourCameraOfCCDs'

    def __init__(self, cam1=[], cam2=[], cam3=[], cam4=[], orientation='horizontal', sizeofcamera=4, subplot_spec=None, sharecolorbar=True, processingsteps=[],  plotingredients=['image', 'time','colorbar'], **kwargs):
        '''

        Parameters
        ----------
        cam1, etc... : various
            a dictionary with keys of 'ccd1', 'ccd2', 'ccd3', 'ccd4'
            that each contain something that can initialize a sequence

        orientation : str
            'horizontal' = camera 2 is right of camera 1
            'vertical'  = camera 2 is below camera 1 (not yet implemented)

        sizeofcamera : float
            What's the size, in inches, to display a single camera?
        '''

        # set up the basic geometry of the main axes
        sizeofcamera = 4
        N = 4
        self.orientation = orientation
        if self.orientation == 'horizontal':
            cols = N
            rows = 1
        IllustrationBase.__init__(self, rows, cols,
                                  figkw=dict(
                                      figsize=(sizeofcamera * cols, sizeofcamera * rows * 1.25)),
                                  hspace=0.02, wspace=0.02,
                                  left=0.05, right=0.95,
                                  bottom=0.1, top=0.9,
                                  subplot_spec=subplot_spec,
                                  sharecolorbar=sharecolorbar)

        # initiate the axes for each camera
        for i in range(rows):
            for j in range(cols):



                # create a CameraFrame for this camera
                n = i * cols + j + 1
                name = 'cam{}'.format(i * cols + j + 1)
                isfullofccds = type(locals()[name]) == dict
                assert(isfullofccds)
                # kludge -- good to check, but this will break if we only have 3/4 cameras

                # make a camera frame to contain these CCDs
                cameraframe = cameras[name](ax=None,
                                            data=None,
                                            illustration=self,
                                            **kwargs)

                # this should be a dictionary of CCDs
                ccds = locals()[name]

                # use an illustration to place a camera in this subplot spec
                thiscameraillustration = CameraOfCCDsIllustration(
                        orientation=orientation,
                        sizeofcamera=sizeofcamera,
                        subplot_spec=self.grid[i,j],
                        camera=name,
                        processingsteps=processingsteps,
                        plotingredients=plotingredients,
                        **ccds)

                # kludge?! this illustration doesn't get plotted when the supersceding one is
                # thiscameraillustration.plotted = {}

                # register the CCD frames (but not the cameras)
                for k in thiscameraillustration.frames.keys():
                    framekey = '{}-{}'.format(name,k)
                    self.frames[framekey] = thiscameraillustration.frames[k]
                    self.frames[framekey].illustration = self

        self._condense_timelabels()



from .imshowFrame import *
from astropy.nddata.utils import Cutout2D
from matplotlib.patches import Rectangle
from .LocalZoomFrame import LocalZoomFrame


class LocalStampFrame(imshowFrame):
    '''
    This frame imshows a Stamp object
    at its appropriate location on
    some source CameraFrame.
    '''

    frametype = 'Stamp'

    def __init__(self, data, source, position=(0, 0), zoom=10, **kwargs):
        '''
        Initialize a LocalStampFrame that can display
        images taken from a zoomed subset of
        a `source` imshowFrame. This is kind
        of like a magnifying class you can
        drop down on an image, but its data
        comes from a single stamp at a time.

        It *doesn't* add a new axis.

        Parameters
        ----------

        data : Stamp
                A StampSequence that contains all its own data.

        source : imshowFrame
                A frame into which this one will be zooming.

        position : (tuple)
                The (x, y) = (col, row) position of the center of zoom.
                (These are in original image coordinates,
                not those that have been transformed for display.)

        '''

        imshowFrame.__init__(self,  data=data, **kwargs)

        self.source = source
        self.source.includes.append(self)

        self.position = position
        self.zoom = zoom

        self.titlefordisplay = '{} | {}'.format(self.frametype, self.position)

    def _get_times(self):
        '''
        Get the available times associated with this frame.
        '''
        return self.data._get_times()

    def _find_timestep(self, time):
        '''
        Get the timestep, by using the source frame.
        '''
        return self.data._find_timestep(time)

    def position_center(self):
        return self.data.col_cent

    def plot(self, timestep=0, **kwargs):

        self.plotted = {}

        # pull out the array to work on
        image, actual_time = self._get_image()

        # pull out the cmap, normalization, and suggested ticks
        cmap, norm, ticks = self.source._cmap_norm_ticks(image)

        # these are in the rotated camera frame
        x, y = self.position
        ysize, xsize = self.data[0].shape

        # set up the extent for the zoomed stamp
        left, right = x - xsize * self.zoom / 2, x + xsize * self.zoom / 2
        bottom, top = y - ysize * self.zoom / 2, y + ysize * self.zoom / 2
        extent = [left, right, bottom, top]

        # figure out the zorder for the stamps
        zooms = [self.illustration.frames[k]
                 for k in self.illustration.frames.keys() if 'stamp' in k]
        zorder = zooms.index(self)

        # draw the imshow
        self.plotted['image'] = self.source.ax.imshow(image,
                                                       extent=extent,
                                                       interpolation='nearest',
                                                       origin='lower',
                                                       norm=norm, cmap=cmap,
                                                       zorder=zorder)

        # Create a Rectangle patch
        rect = Rectangle((left, bottom), xsize * self.zoom, ysize * self.zoom,
                         linewidth=1, edgecolor='black', facecolor='none',
                         zorder=zorder + 0.5, clip_on=True)

        # Add the patch to the Axes
        self.plotted['boxonzoom'] = self.source.ax.add_patch(rect)

        # make the axes clip
        self.source.ax.set_clip_on(True)

        # add a box to the source image (cutout must have ben created, if plot has happened)
        #boxonoriginal = self.cutout.plot_on_original(ax=self.source.ax, clip_on=True)

        # make sure there's a colorbar
        self.source._ensure_colorbar_exists(self.plotted['image'])

        # keep track of the current timestep
        self.currenttimestep = timestep

    def update(self, time):
        '''
        Update this frame to a particular time (for use in animations).
        '''
        # update the data, if we need to
        timestep = self._find_timestep(time)
        image, actual_time = self._get_image(time)
        if image is None:
            return

        if timestep != self.currenttimestep:
            self.plotted['image'].set_data(image)
            timestring = self.source._timestring(actual_time)
            self.source.plotted['time'].set_text(timestring)



from __future__ import print_function

'''
If something inherits from Talker, then we can print
text to the terminal in a relatively standard way.
'''

import textwrap
import numpy as np
import pprint
import sys
if sys.version_info[0] < 3:
    input = raw_input

shortcuts = None
line = 100

class Talker:
    '''
    Objects the inherit from Talker have "mute" and "pithy" attributes,
        a report('uh-oh!') method that prints when unmuted,
        a speak('yo!') method that prints only when unmuted and unpithy,
        and an input("what's up?") method that takes input from the prompt.
    '''

    def __init__(self, nametag=None, mute=False, pithy=False, line=line, prefixformat="{0:>16}"):
        self._mute = mute
        self._pithy = pithy
        self._line = line
        self._prefixformat = prefixformat
        if nametag is None:
            self.nametag = self.__class__.__name__.lower()
        else:
            self.nametag = nametag
        self.nametag = self.nametag.replace('_', '-')

    def speak(self, string='', level=0, progress=False):
        '''If verbose=True and terse=False, this will print to terminal. Otherwise, it won't.'''
        if self._pithy == False:
            self.report(string=string, level=level, progress=progress)

    def warning(self, string='', level=0):
        '''If verbose=True and terse=False, this will print to terminal. Otherwise, it won't.'''
        self.report(string, level, prelude=':-| ')

    def input(self, string='', level=0, prompt='(please respond) '):
        '''If verbose=True and terse=False, this will print to terminal. Otherwise, it won't.'''
        self.report(string, level)
        return input("{0}".format(self._prefix + prompt))

    def report(self, string='', level=0, prelude='', progress=False, abbreviate=True):
        '''If verbose=True, this will print to terminal. Otherwise, it won't.'''
        if self._mute == False:
            self._prefix = prelude + \
                '{spacing}[{name}] '.format(
                    name=self.nametag, spacing=' ' * level)
            self._prefix = self._prefixformat.format(self._prefix)
            equalspaces = ' ' * len(self._prefix)
            toprint = string + ''
            if abbreviate:
                if shortcuts is not None:
                    for k in shortcuts.keys():
                        toprint = toprint.replace(k, shortcuts[k])

            if progress:
                end = '\r'
            else:
                end = '\n'

            print(self._prefix + '\n'.join(textwrap.wrap(toprint, width=self._line - len(self._prefix))).replace('\n', '\n' + equalspaces), end=end)


    def summarize(self):
        '''Print a summary of the contents of this object.'''

        self.speak('Here is a brief summary of {}.'.format(self.nametag))
        s = '\n' + pprint.pformat(self.__dict__)
        print(s.replace('\n', '\n' + ' ' * (len(self._prefix) + 1)) + '\n')



from .IllustrationBase import *
from .StampsIllustration import *
from .FourCameraIllustration import *
from .FourCameraOfCCDsIllustration import *
from .CameraIllustration import *
from .imshowIllustration import *
from .CameraOfCCDsIllustration import *
from .CCDIllustration  import *
from .SingleCameraWithZoomIllustration import *
from .GenericIllustration import GenericIllustration
from .SideBySideIllustration import SideBySideIllustration
from ..zoom import *



from ...imports import *

from astropy.modeling import models, fitting
from matplotlib.patches import Ellipse
from astropy.table import Table, vstack
from astropy.stats import SigmaClip
from photutils import Background2D, MedianBackground, SExtractorBackground


okcmap = one2another('red', 'red', alphabottom=0.3, alphatop=0)

def create_test_image(gap=3, N=25):
    '''
    Create one random stitched camera, with a Gaussian-ish background to it.
    '''

    # generate a random model
    p_generate = (models.Const2D(amplitude=10**np.random.uniform(0, 1)) +
               models.Gaussian2D(amplitude=10**np.random.uniform(.5, 2),
                                   x_mean=np.random.uniform(-N, N),
                                   y_mean=np.random.uniform(-N, N),
                                   x_stddev=np.random.uniform(2, N),
                                   y_stddev=np.random.uniform(2, N),
                                   theta=np.random.uniform(0, 2*np.pi)))

    # create fake x and y arrays
    x, y = np.meshgrid(np.arange(-N-gap, N+gap+1), np.arange(-N-gap, N+gap+1))

    # create the model
    perfect = p_generate(x, y)
    z = np.random.normal(perfect, 1)

    # find the gaps, replace them with zeros, and mark them as bad
    bad = (np.abs(x) < gap) | (np.abs(y) < gap)
    z[bad] = 0
    ok = bad == False

    return x, y, z, ok, p_generate

def showimage(x, y, z, ok, **kw):
    '''
    Use imshow to show all four CCDs.
    '''
    imshowkw = dict(origin='lower',
                    extent=[np.min(x), np.max(x), np.min(y), np.max(y)],
                    interpolation='nearest', **kw)

    # mask out the neglected pixels
    maskkw = dict(**imshowkw)
    maskkw['cmap'] = okcmap
    maskkw['vmin'] = 0
    maskkw['vmax'] = 1
    plt.imshow(ok, zorder=100, **maskkw)

    # show the image
    plt.imshow( z, zorder=-100, **imshowkw)
    plt.axis('scaled')

def bin_image(image, binning=100, visualize=False):
    '''
    Bin an image down by a given factor,
    by taken the means of squares of pixels.

    Parameters
    ----------
    image : a

    '''
    rows = image.shape[0]//binning
    binnedoverrow = image[:rows*binning, :].reshape(rows, binning, image.shape[1]).mean(1)
    cols = image.shape[1]//binning
    binned = binnedoverrow[:,:cols*binning].reshape(binnedoverrow.shape[0], cols, binning).mean(2)

    if visualize:
        plt.imshow(binned)
        plt.colorbar()

    return binned

def guess_2d_gaussian(x, y, z, ok=None):
    '''
    Make a guess to initialize the baseline + 2D Gaussian model,
    based on medians and weighted moments of the image.
    '''

    if ok == None:
        ok = np.ones_like(z).astype(np.bool)

    # estimate a baseline flux, from the median of the good pixels
    baseline_guess = np.median(z[ok])

    # do a *veyr* coarse subtraction, to focus on just the blob
    crudelysubtracted = np.maximum((z-baseline_guess), 0)
    amplitude_guess = np.sum(crudelysubtracted)


    def moment(q):
        '''
        Take the weighted moment of a quantity.
        '''
        return np.sum(q[ok]*crudelysubtracted[ok])/np.sum(crudelysubtracted[ok])

    # calculate flux-weighted centroids of the blob
    x_guess = moment(x)
    y_guess = moment(y)

    # calculate flux-weighted widths of the blob
    x_width_guess = np.sqrt(moment((x - x_guess)**2))
    y_width_guess = np.sqrt(moment((y - y_guess)**2))



    # create an initial model, with the initial guesses
    initial = (  models.Const2D(amplitude=baseline_guess) +
                models.Gaussian2D(amplitude=amplitude_guess,
                               x_mean=x_guess, y_mean=y_guess,
                               x_stddev=x_width_guess, y_stddev=y_width_guess,
                               theta=0.0))

    return initial

def fit_model(x, y, z, ok=None, model=None, imagebinning=1, **kw):
    '''
    Fit the scattered light in a list of images (with x, y, z) for each.

    Parameters
    ----------
    x, y, z : 2D arrays
        The x and y values of each pixel (kind of boring images), and the z flux of the image.
    ok : 2D array
        Mask set to 1 where data are OK, 0 where they should be ignored.
    model : an astropy Model
        The initialized model to fit.
    imagebinning : int
        By how many pixels should the images be binned?
        This substantially speeds up fitting.
    p_init : an
        Initial guesses for the parameters of Gaussian2D.

    '''


    # construct the fitter (using LM)
    fit_p = fitting.LevMarLSQFitter()

    # run the fit (skipping some warnings)
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')

        # if desired, bin the data
        if imagebinning == 1:
            bx, by, bz, bok = x, y, z, ok
        else:
            bx = bin_image(x, binning=imagebinning)
            by = bin_image(y, binning=imagebinning)
            bz = bin_image(z, binning=imagebinning)
            # (exclude bins that are only partially OK)
            bok = bin_image(ok, binning=imagebinning) == 1

        # run the fit, starting with initial model and comparing to the data
        p = fit_p(model, bx[bok], by[bok], bz[bok])

    return p

def plot_2d_gaussian(p, color='darkorchid', **kw):
    '''
    Using an astropy model Gaussian2D, plot it on an image.
    **kw go into Ellipse definition
    '''

    # plot the ellipse on each image
    baseline, gaussian = p
    e = dict(zip(gaussian.param_names, gaussian.parameters))
    howmanysigma = 2
    ellipse = Ellipse(xy=(e['x_mean'], e['y_mean']),
                      width=howmanysigma*e['x_stddev'],
                      height=howmanysigma*e['y_stddev'],
                      angle=e['theta']*180/np.pi,
                      linewidth=3, facecolor='none',
                      edgecolor=color, alpha=0.5, **kw)
    plt.gca().add_artist(ellipse)


def plot_fit(x, y, z, ok, models=None, colors=None, cmap='gray', **kw):
    '''
    Plot images along with their 2D Gaussian model.

    x, y, z : the two coordinate images and the flux image
    p : the model 2D Gaussian to plot, using an ellipse as summary

    **kw goes into showimage
    '''

    # set the scale for the image to plot
    vmin, vmax = np.percentile(z[ok], [1, 99])

    # plot the data with the best-fit model
    fi, ax = plt.subplots(1, 3, figsize=(10, 2.5), sharex=True, sharey=True)


    fitted = models[0]
    plt.sca(ax[0])
    showimage(x, y, z, ok, vmin=vmin, vmax=vmax, cmap=cmap, **kw)

    for m, c in zip(models, colors):
        plot_2d_gaussian(m, color=c)
    plt.axis('off')
    textkw = dict(color='white', alpha=0.5, weight='bold')
    plt.text(0.05, 0.05, 'data', transform=plt.gca().transAxes, **textkw)

    plt.sca(ax[1])
    showimage(x, y, fitted(x,y), ok, vmin=vmin, vmax=vmax, cmap=cmap, **kw)
    for m, c in zip(models, colors):
        plot_2d_gaussian(m, color=c)
    plt.axis('off')
    plt.text(0.05, 0.05, 'model', transform=plt.gca().transAxes, **textkw)


    plt.sca(ax[2])
    showimage(x, y, z-fitted(x,y), ok, vmin=vmin, vmax=vmax, cmap=cmap, **kw)
    for m, c in zip(models, colors):
        plot_2d_gaussian(m, color=c)
    plt.axis('off')
    plt.text(0.05, 0.05, 'residual', transform=plt.gca().transAxes, **textkw)

    plt.colorbar(ax=ax)

def load_camera(filename):
    '''
    Load a stitched full camera image.

    Parameters
    ----------
    filename : string
        The path to a FITS file containing faus' background images.
        (for example, "orbit-4104/bkg_cam1_spm1_1208878859.cal.fits")

    Returns
    -------
    x, y, z, ok :
        The x coordinate, the y coordinate, the image, mask of the good data.

    '''


    hdu = fits.open(filename)
    z = hdu[0].data
    rows, columns = z.shape
    x1d = np.arange(columns) - columns/2
    y1d = np.arange(rows) - rows/2
    x, y = np.meshgrid(x1d, y1d)

    # faus set the chip gaps to 0
    ok = z != 0

    return x, y, z, ok

def test(**kw):
    '''
    Test out the model fitting on a test image.
    '''

    # create a fake test image
    x, y, z, ok, truth = create_test_image(**kw)

    # throw out the corners (flux values of exactly 0 are ignored)
    r = np.sqrt(x**2 + y**2)
    incircle = r < np.max(x)
    ok *= incircle
    #z[incircle == False] = 0

    # make an initial guess model
    initial = guess_2d_gaussian(x,y,z)

    # fit it (binning by 100)
    fitted = fit_model(x, y, z, ok, model=initial)

    #plot_fit(xsmall, ysmall, zsmall, truth)
    plot_fit(x, y, z, ok, models=[fitted, initial, truth], colors=['darkorchid', 'gray', 'hotpink'])


def remove_stars_from_image(x, y, z, ok, box=150, filter=3, visualize=True, label=''):

    withstars = z
    sigma_clip = SigmaClip(sigma=3., iters=10)
    bkg_estimator = MedianBackground()#SExtractorBackground()#
    bkg = Background2D(withstars, box_size=(box,box), filter_size=(filter,filter), mask=ok==False,
                        sigma_clip=sigma_clip, bkg_estimator=bkg_estimator)


    withoutstars = bkg.background*ok
    if visualize:

        fi, ax = plt.subplots(1, 3, figsize=(10,2.5), sharex=True, sharey=True)

        kw = dict(vmin=np.percentile(withoutstars, 2), vmax=np.percentile(withstars, 98), cmap='gray')
        textkw = dict(color='white', alpha=0.5, weight='bold')

        plt.sca(ax[0])
        showimage(x, y, z, ok, **kw)
        plt.text(0.05, 0.05, 'original', transform=plt.gca().transAxes, **textkw)
        plt.axis('off')

        plt.sca(ax[1])
        showimage(x, y, withoutstars, ok, **kw)
        plt.text(0.05, 0.05, 'background', transform=plt.gca().transAxes, **textkw)
        plt.axis('off')

        plt.sca(ax[2])
        showimage(x, y, withstars - withoutstars, ok, **kw)
        plt.text(0.05, 0.05, 'subtracted', transform=plt.gca().transAxes, **textkw)
        plt.axis('off')

        plt.suptitle(label)
        plt.colorbar(ax=ax)

    return withoutstars


def fit_camera(filename,
               binby=200,
               constrain_theta='clockhands',
               constrain_size=(100, 3000),
               visualize=True,
               label=None,
               removestars=False, **kw):
    '''
    Fit an image from a full camera.

    Parameters
    ----------

    filename : str
        Path to the FITS file for a full-camera stitched image.
    binby : int
        Factor by which to bin the image (speeds up fitting).
    constrain_theta : str or None
        'clockhands' means the angle rotates with location in the field
        None means the angle of the Gaussian is unconstrained
    constrain_size : 2-element tuple
        The min and max values for the widths of the blobs
    label : str
        How should files be saved?

    Returns:
    table : an astropy Table
        a single-row table of parameters for diffuse light fit

    '''

    if label is None:
        label = os.path.basename(filename).replace('.fits', '')

    fitfilename = 'fits/fit_{}.txt'.format(label)
    if os.path.exists(fitfilename):
        print("{} already exists -- not refitting.".format(fitfilename))
        row = ascii.read(fitfilename)
        return row

    # load an image from a FITS file
    x, y, z, ok = load_camera(filename)

    # throw out the corners
    r = np.sqrt(x**2 + y**2)
    ok *= r < np.max(x)

    # do we need to remove the stars from this image?
    if removestars:
        z = remove_stars_from_image(x, y, z, ok, visualize=visualize, label=label, **kw)
        if visualize:
            plotfilename = os.path.join('plots', 'background_{}.png'.format(label))
            plt.savefig(plotfilename, dpi=400)

    # make an initial guess model
    initial = guess_2d_gaussian(x,y,z)
    base, gauss = initial
    # should we make sure theta is is set by the position?
    if constrain_theta == 'clockhands':
        def tiedtheta(initial):
            '''
            Define theta from the clock position in the image.
            '''
            b, g = initial
            return np.arctan2(g.y_mean, g.x_mean)
        gauss.theta.tied = tiedtheta

    # should we set a typical size for the blob?
    if constrain_size:
        gauss.x_stddev.bounds = constrain_size
        gauss.y_stddev.bounds = constrain_size

    # fit it (binning by 100)
    fitted = fit_model(x, y, z, ok, model=initial, imagebinning=binby)


    if visualize:

        #plot_fit(xsmall, ysmall, zsmall, truth)
        plot_fit(x, y, z, ok, models=[fitted, initial], colors=['darkorchid', 'gray'])
        plt.suptitle(label)
        plotfilename = os.path.join('plots', 'fit_{}.png'.format(label))
        plt.savefig(plotfilename, dpi=400)


    b, g = fitted
    binitial, ginitial = initial
    imtype, camera, spm, time = label.split('_')#os.path.basename(filename).split('.')[0].split('_')
    d = dict(baseline=b.amplitude.value,
             amplitude=g.amplitude.value,
             x=g.x_mean.value,
             y=g.y_mean.value,
             radius=(g.x_mean.value**2 + g.y_mean.value**2),
             theta=g.theta.value,
             radial_width=g.x_stddev.value,
             theta_width=g.y_stddev.value,
             moment_x=ginitial.x_mean.value,
             moment_y=ginitial.y_mean.value,
             imtype=imtype,
             camera=camera,
             spm=spm,
             spacecrafttime=time,
             filename=os.path.basename(filename))

    row = Table([d], names=['baseline', 'amplitude', 'x', 'y', 'radius', 'theta', 'radial_width', 'theta_width', 'moment_x', 'moment_y', 'imtype', 'camera', 'spm', 'spacecrafttime', 'filename'])

    row.write(fitfilename, format='ascii.fixed_width', delimiter='|', bookend=False, overwrite=True)

    return row



from .visualize import *
from ...imports import *
from ...postage import *

def bin_jitter(lc, binwidth=30.0/60./24, robust=False):
    '''
    Bin the jitter to a useful cadence.
    '''

    c, r = lc.centroid_col, lc.centroid_row

    #bx, by, be = binto(lc.time, np.sqrt(c**2 + r**2), binwidth=binwidth, sem=False, robust=True)
    #plt.bar(bx, be, width=binwidth, alpha=0.3, label='radial')

    time, ry, re = binto(lc.time, r, binwidth=binwidth, sem=False, robust=False)
    _, cy, ce = binto(lc.time, c, binwidth=binwidth, sem=False, robust=False)


    centroid_col = cy - np.nanmedian(cy)
    centroid_row = ry - np.nanmedian(ry)
    intrajitter_col = ce
    intrajitter_row = re
    return time, centroid_col, centroid_row, intrajitter_col, intrajitter_row


def change_pipeline_aperture(tpf, aperture, backgroundaperture):
    tpf.hdu[-1].data = np.maximum(3*aperture, 2*backgroundaperture)

def define_apertures(tpf,  apertureradius=3, searchradius=5, backgroundpercentile=20, visualize=True):

    # make a stacked image of this stamp
    image = np.mean(tpf.flux, 0)

    # identify which pixels we want to call the background
    backgroundaperture = image <= np.percentile(image, backgroundpercentile)
    image -= np.mean(image[backgroundaperture])

    # create a search circle at the center of the stamp
    col, row = np.meshgrid(np.arange(image.shape[1]), np.arange(image.shape[0]))
    rcenter, ccenter = np.mean(row), np.mean(col)
    incircle = np.sqrt((row-rcenter)**2 + (col-ccenter)**2) < searchradius

    # find the flux-weighted centroid within that search circle
    rcentroid = np.sum(row*image*incircle)/np.sum(image*incircle)
    ccentroid = np.sum(col*image*incircle)/np.sum(image*incircle)

    # define the aperture
    aperture = np.sqrt((row-rcentroid)**2 + (col-ccentroid)**2) < apertureradius

    # actually apply the changes to the tpf
    change_pipeline_aperture(tpf, aperture, backgroundaperture)

    return aperture, backgroundaperture

def subtract_background(tpf):

    backgroundaperture = tpf.hdu[-1].data == 2
    background1d = np.median(tpf.hdu[1].data['RAW_CNTS'][:, backgroundaperture], 1)

    # store the background in the FITS
    tpf.hdu[1].data['FLUX_BKG'][:,:,:] = background1d[:, np.newaxis, np.newaxis]
    tpf.hdu[1].data['FLUX'] = tpf.hdu[1].data['RAW_CNTS'] - tpf.hdu[1].data['FLUX_BKG']

    #if visualize:
    #    pass
    #    #ax = tpf.to_lightcurve().plot(normalize=False, label='target')
    #    plt.plot(background1d*np.sum(tpf.pipeline_mask), label='background')
    #    plt.plot(np.sum(tpf.hdu[1].data['RAW_CNTS'][:, tpf.pipeline_mask], 1), label='total')

        #plt.legend()
        #plt.show()

def stamps_to_tpfs(files, tpfdirectory='.', **kw):
    '''
    Take a list of .npy stamp files,
    and convert them to TPFs.
    '''
    for f in files:
        print('converting {} into a proper TPF'.format(f))
        s = Stamp(f)
        tpf = EarlyTessTargetPixelFile.from_stamp(s)
        print('saving {} to {}'.format(tpf, tpfdirectory))
        tpf.to_fits(directory=tpfdirectory, **kw)


def photometer(tpf, **kw):

    # set the apertures
    define_apertures(tpf, **kw)

    # subtract the background from the flux array
    subtract_background(tpf)

    lc = tpf.to_lightcurve('pipeline')
    return lc

def calculate_differences(tpf, **kw):
    '''
    Make a copy of a TPF that is populated with time-differences.
    '''

    differenced = copy.deepcopy(tpf)

    differenced.hdu[1].data['FLUX'][:-1,:,:] = np.diff(tpf.flux, axis=0)
    differenced.hdu[1].data['FLUX'][-1,:,:] = 0

    #differenced.hdu[1].data['FLUX'][0,:,:] = 0#np.nan

    return differenced

def calculate_gradients(image, visualize=False):
    '''
    Make the 2D image gradients of the image.

    Parameters
    ----------

    image : 2D array
        For example, a median-stacked image.

    Returns
    -------

    dMdx_forward, dMdy_forward, dMdx_forward, dMdy_backward : 2D arrays
        Each the same size as the original image,
        containing the image gradient components.
    '''
    dMdx_forward, dMdx_backward, dMdy_forward, dMdy_backward = [np.zeros_like(image) for _ in range(4)]
    dMdx_forward[:,:-1] = np.diff(image, axis=1)
    dMdx_backward[:,1:] = np.diff(image, axis=1)
    dMdy_forward[:-1,:] = np.diff(image, axis=0)
    dMdy_backward[1::] = np.diff(image, axis=0)

    if visualize:
        fi, ax = plt.subplots(4, 2, sharex=True, sharey=True, figsize=(3,7))
        grady, gradx = np.gradient(image)# np.diff(image[::-1], axis=1)[::-1]
        kw = dict(origin='lower')
        ax[0,0].imshow(image, **kw)
        ax[0,1].set_visible(False)
        ax[1,0].imshow(dMdx_forward, **kw)
        ax[2,0].imshow(dMdx_backward, **kw)
        ax[3,0].imshow(gradx, **kw)

        ax[1,1].imshow(dMdy_forward, **kw)
        ax[2,1].imshow(dMdy_backward, **kw)
        ax[3,1].imshow(grady, **kw)

        ax[1,0].set_ylabel('forward')
        ax[2,0].set_ylabel('backward')
        ax[3,0].set_ylabel('threepix')

        ax[3,0].set_xlabel('x'); ax[3,1].set_xlabel('y')
    return dMdx_forward, dMdy_forward, dMdx_forward, dMdy_backward




# NEXT STEPS
# -compare light curves with/without CRM
# -plot and do photometry on differences between +/- CRM
#



#!/usr/bin/env python

# The template for this setup.py came from Tim Morton,
# who I understand took it from Dan F-M. And then Geert
# Barentsen and Christina Hedges helped explain a few
# more neat tips. Thanks all!


import os
import sys
from setuptools import setup, find_packages


# a little kludge to get the version number from __version__
exec(open('illumination/version.py').read())

# running `python setup.py release` from the command line will post to PyPI
if "release" in sys.argv[-1]:
    os.system("python setup.py sdist")
    # uncomment the next line to test out on test.pypi.com/project/tess-zap
    #os.system("twine upload --repository-url https://test.pypi.org/legacy/ dist/*")
    os.system("twine upload dist/*")
    os.system("rm -rf dist/henrietta*")
    sys.exit()

setup(name = "illumination",
    version = __version__,
    description = "Python tools for visualizing astronomical images, particularly from the NASA TESS mission.",
    long_description = "", # need to add
    author = "Zach Berta-Thompson",
    author_email = "zach.bertathompson@colorado.edu",
    url = "https://github.com/zkbt/illumination",
    packages = find_packages(),
    package_data = {'illumination':[]},
    include_package_data=False,
    scripts = [],
    classifiers=[
      'Intended Audience :: Science/Research',
      'Programming Language :: Python',
      'Topic :: Scientific/Engineering :: Astronomy'
      ],
    install_requires=['numpy', 'matplotlib', 'scipy', 'astropy', 'lightkurve>=1.0b21'],
    zip_safe=False,
    license='MIT',
)




from IPython.display import HTML

def display(filename):
    '''
    Display files of different types in a jupyter notebook.
    '''

    if '.mp4' in filename:
        template = """<video width="80%" controls loop>
                      <source src="{}" type="video/mp4" />
                      </video>"""
    elif '.pdf' in filename:
        template = """<iframe src={} width=80%></iframe>'.format(filename)"""
    else:
        template = """<img width="80%" src={}>"""

    return HTML(template.format(filename))



from __future__ import print_function

'''
These define a common set of imports, which can be used by other modules
throughout the entire package.
'''

import os
import copy
import subprocess
import glob
import shutil
import warnings
import numpy as np
import matplotlib.pyplot as plt
import scipy.stats

import matplotlib.animation as ani
import matplotlib.gridspec as gs
import matplotlib.colors as co

from astropy.io import fits, ascii
from astropy.time import Time
from astropy.stats import mad_std, sigma_clip
import astropy.units as u

from .talker import Talker


'''
FIXME -- Move these to a separate utilities.py file?
'''

def mkdir(path):
    '''
    A mkdir that doesn't complain if it already exists.
    '''
    try:
        os.mkdir(path)
        print("made {}".format(path))
    except:
        pass


def mad(x):
    '''
    Returns the median absolute deviation from the median,
                    a robust estimator of a distribution's width.

                    For a Gaussian distribution, sigma~1.48*MAD.
    '''
    med = np.nanmedian(x)
    return np.nanmedian(np.abs(x - med))


def shared_directory(files, verbose=False):
    '''
    Find the shared base directory amongst a list of files.

    Parameters
    ----------

    files : list
            A list of filenames.

    Returns
    -------

    shared : str
            A filepath that is the shared across all the files.
    '''

    for i in range(len(files[0])):
        shared = files[0][:i + 1]
        if verbose:
            print('"{}"'.format(shared))
        for f in files:
            if f[:i + 1] != shared:
                if verbose:
                    print('Huzzah! "{}" is shared across {} files.'.format(
                        shared[:-1], len(files)))
                return shared[:-1]


def binto(x=None, y=None, yuncertainty=None,
          binwidth=0.01,
          test=False,
          robust=True,
          sem=True,
          verbose=False):
    '''Bin a timeseries to a given binwidth,
            returning both the mean and standard deviation
                    (or median and approximate robust scatter).'''

    if test:
        n = 1000
        x, y = np.arange(n), np.random.randn(n)  # - np.arange(n)*0.01 + 5
        bx, by, be = binto(x, y, binwidth=20, robust=robust, sem=sem)
        plt.figure('test of craftroom.binto')
        plt.cla()
        plt.plot(x, y, linewidth=0, markersize=4,
                 alpha=0.3, marker='.', color='gray')
        plt.errorbar(bx, by, be, linewidth=0, elinewidth=2, capthick=2,
                     markersize=10, alpha=0.5, marker='.', color='blue')
        return

    bias = np.nanmean(y)
    min, max = np.min(x[np.isfinite(x)]), np.max(x[np.isfinite(x)])
    bins = np.arange(min, max + binwidth, binwidth)
    count, edges = np.histogram(x, bins=bins)
    sum, edges = np.histogram(x, bins=bins, weights=y)

    if yuncertainty is not None:
        count, edges = np.histogram(x, bins=bins)
        numerator, edges = np.histogram(
            x, bins=bins, weights=y / yuncertainty**2)
        denominator, edges = np.histogram(
            x, bins=bins, weights=1.0 / yuncertainty**2)
        mean = numerator / denominator
        std = np.sqrt(1.0 / denominator)
        error = std
        if False:
            for i in range(len(bins) - 1):
                print(bins[i], mean[i], error[i], count[i])
            a = raw_input('???')
    else:
        if robust:
            n = len(sum)
            mean, std = np.zeros(n) + np.nan, np.zeros(n) + np.nan
            for i in range(n):
                inbin = (x > edges[i]) * (x <= edges[i + 1])
                mean[i] = np.median(y[inbin])
                std[i] = 1.48 * mad(y[inbin])
        else:
            if yuncertainty is None:
                mean = sum.astype(np.float) / count
                sumofsquares, edges = np.histogram(x, bins=bins, weights=y**2)

                # *np.sqrt(count.astype(np.float)/np.maximum(count-1.0, 1.0))
                std = np.sqrt((sumofsquares.astype(np.float) -
                               count * mean**2) / (count - 1))
                # plt.plot(count*mean**2)
                # plt.plot(sumofsquares)
                # plt.show()

        if sem:
            error = std / np.sqrt(count)
        else:
            error = std

    x = 0.5 * (edges[1:] + edges[:-1])
    return x, mean, error

    if yuncertainty is not None:
        print("Uh-oh, the yuncertainty feature hasn't be finished yet.")

    if robust:
        print("Hmmm...the robust binning feature isn't finished yet.")


def name2color(name):
    '''
    Return the 3-element RGB array of a given color name.
    '''
    return co.hex2color(co.cnames[name])


def one2another(bottom='white', top='red', alphabottom=1.0, alphatop=1.0, N=256):
    '''
    Create a cmap that goes smoothly (linearly in RGBA) from "bottom" to "top".
    '''
    rgb_bottom, rgb_top = name2color(bottom), name2color(top)
    r = np.linspace(rgb_bottom[0], rgb_top[0], N)
    g = np.linspace(rgb_bottom[1], rgb_top[1], N)
    b = np.linspace(rgb_bottom[2], rgb_top[2], N)
    a = np.linspace(alphabottom, alphatop, N)
    colors = np.transpose(np.vstack([r, g, b, a]))
    cmap = co.ListedColormap(colors, name='{bottom}2{top}'.format(**locals()))
    return cmap



'''Generate TESS pixel lightcurve cubes with dimensions (xpix)x(ypix)x(time).'''
from ..imports import *
from .stackers import Central, Sum

timeaxis = 0
class Cube(Talker):
	'''
	Cube to handle simulated postage stamp pixel light curves;
			has dimensions of (xpixels, ypixels, time).
	'''
	_savable = ['photons', 'temporal', 'static', 'spatial']
	identifier = ''
	label = ''
	colorbarlabelfordisplay = ''

	def __init__(self, photons,
					   cadence=2, # what cadence is this cube (in s)
					   time=None, # array of times for the cube
					   temporal={}, # temporal variables
					   spatial={}, # spatial variables
					   static={}, # static variables
					   ):
		'''
		Initialize a cube object.

		Parameters:
		-----------

		array : (xpixels, ypixels, time) array
		cadence?
		subexposurecadence?
		'''

		# decide whether or not this Cube is chatty
		Talker.__init__(self)

		# create an observation
		self.cadence = cadence

		# store the photons as a 3D array
		self.photons = photons

		# set shapes
		self.shape = self.photons.shape
		self.n = self.shape[0]
		self.ypixels = self.shape[1]
		self.xpixels = self.shape[2]

		# store the other diagnostics (including time)
		self.temporal = temporal

		# these times are all in seconds! (probably GPS?)
		try:
			self.time = self.temporal['TIME']
		except KeyError:
			self.speak('making up imaginary times')
			self.time = Time('2018-01-01 00:00:00.000').gps + np.arange(self.n)*self.cadence
		self.spatial = spatial
		self.static = static

		# make up a fake time axis
		# default for plotting
		self.consider()
		# create a dictionary to store a bunch of summaries
		self.summaries = {}

	def __repr__(self):
		'''
		How do we print out this cube?
		'''
		return '<Cube | {} | {}s>'.format(self.shape, self.cadence)


	def stack(self, cadence, strategy=Central(n=10), plot=False):
		'''
		Stack together 2-second exposures, using some cosmic strategy.
		'''

		# by how many subexposures do we need to bin this one?
		nsubexposures = np.int(cadence/self.cadence)

		# loop over the x and y pixels
		self.speak('binning {0} cube by {1} subexposures'.format(self.shape, nsubexposures))

		# make the stacked array
		array = strategy(self.photons, nsubexposures)

		temporal = {}
		for k in self.temporal.keys():
			temporal[k] = strategy.average1d(self.temporal[k], nsubexposures)

		static = dict(**self.static)
		# return the stacked array

		binned = Cube(array, cadence, temporal=temporal, spatial=self.spatial, static=self.static)

		return binned

	def consider(self, what='counts'):
		'''
		Decide what array to visualize in plots.

		Parameters
		----------

		what : str
			Describe what should be visualized.

		'''

		if what == 'photons':
			self.todisplay = self.photons
			#def label(i=0):
			#	return '{}s (photons)'.format(self.temporal['TIME'][i])
			self.colorbarlabelfordisplay = 'counts'

		if what == 'counts':
			# kludge! (because don't have calibrated data yet!)
			self.todisplay = self.photons
			#def label(i=0):
			#	return '{}s (counts)'.format(self.temporal['TIME'][i])
			self.colorbarlabelfordisplay = 'counts'

		if what == 'differences':
			self.todisplay = np.diff(self.photons, axis=0)
			#def label(i=0):
			#	return '{}s - {}s (counts)'.format(self.temporal['TIME'][i+1], self.temporal['TIME'][i])
			self.colorbarlabelfordisplay = 'difference (counts)'

	@property
	def titlefordisplay(self):
		'''
		What to be displayed as a title for a plot
		(this should probably be moved to a Frame instead?)
		'''
		return '{}\n{}'.format(self.identifier, self.label)

	def imshow(self, timestep=0, nsigma=0.5):
		'''
		Make an imshow of a single frame of the cube.
		'''

		a = self.todisplay
		vmin, vmax = np.percentile(a, [1,99])
		if (self.todisplay < 0).any():
			scale = np.maximum(np.abs(vmin), np.abs(vmax))
			vmin, vmax = -scale, scale

			sigma = mad(self.todisplay)
			norm = SymLogNorm(nsigma*sigma, linscale=2, vmin=vmin, vmax=vmax)
			cmap = 'RdBu'
			ticks = [vmin, -nsigma*sigma, 0, nsigma*sigma, vmax]
		else:
			norm = LogNorm(vmin=vmin, vmax=vmax)
			cmap = 'Blues'
			ticks = [vmin, vmax]

		# point at a particular image
		image = self.todisplay[timestep, :,:]

		plotted = plt.imshow(image, interpolation='nearest', origin='lower', norm=norm, cmap=cmap)
		plt.title(self.titlefordisplay)
		plt.axis('off')
		colorbar = plt.colorbar(plotted, orientation='horizontal', label=self.colorbarlabelfordisplay(0), fraction=0.04, pad=0.02, ticks=ticks)
		colorbar.ax.set_xticklabels(['{:.0f}'.format(v) for v in ticks])
		colorbar.outline.set_visible(False)

		return plotted, colorbar


	def movie(self, filename='test.gif', fps=30, **kw):
		self.speak('displaying {} exposures of the pixel cube'.format(self.n))

		plt.clf()
		plotted, colorbar = self.imshow(**kw)

		if '.mp4' in filename:
			writer = ani.writers['ffmpeg'](fps=fps)
		else:
			try:
				writer = ani.writers['pillow'](fps=fps)
			except (RuntimeError, KeyError):
				writer = ani.writers['imagemagick'](fps=fps)

		fig = plt.gcf()

		# the "with" construction is a little confusing, but feel free to copy and paste this
		with writer.saving(fig, filename, fig.get_dpi()):

			for i in range(self.todisplay.shape[0]):
				print(i, self.n)

				# update the data to match this frame
				plotted.set_data(self.todisplay[i, :,:])
				colorbar.set_label(self.colorbarlabelfordisplay(i))
				# save this snapshot to a movie frame
				writer.grab_frame()

	def cubify(self, image):
		'''
		Slightly reshape an image by adding an extra dimension,
		so it can be cast into operations on the whole cube.
		'''
		return image.reshape(1, self.xpixels, self.ypixels)

	def median(self, which='photons'):
		'''
		The median image.
		'''
		array = self.__dict__[which].astype(np.float64)
		key = 'median'
		try:
			self.summaries[key+which]
		except:
			self.summaries[key+which] = np.median(array, timeaxis)
		return self.summaries[key+which]

	def mean(self, which='photons'):
		'''
		The mean image.
		'''
		array = self.__dict__[which].astype(np.float64)
		key = 'mean'
		try:
			self.summaries[key+which]
		except:
			self.summaries[key+which] = np.mean(array, timeaxis)
		return self.summaries[key+which]


	def mad(self, which='photons'):
		'''
		The median of the absolute deviation image.
		'''
		array = self.__dict__[which].astype(np.float64)
		key = 'mad'
		try:
			self.summaries[key+which]
		except:
			self.summaries[key+which] = np.median(np.abs(array - self.cubify(self.median(which))), timeaxis)
		return self.summaries[key+which]

	def std(self, which='photons'):
		'''
		The standard deviation image.
		'''
		array = self.__dict__[which].astype(np.float64)
		key = 'std'
		try:
			self.summaries[key+which]
		except:
			self.summaries[key+which] = np.std(array, timeaxis)
		return self.summaries[key+which]

	def sigma(self, which='photons', robust=True):
		'''
		The sigma of the image, using either a robust MAD or not.
		'''
		if robust:
			return 1.4826*self.mad(which)
		else:
			return self.std(which)

	def oneWeirdTrickToTrimCosmics(self, threshold=4):
		'''
		To be used for simulating ground-based cosmics removal.

		(this assumes *zero* jitter)
		'''
		shape = np.array(self.photons.shape)
		shape[timeaxis] = 1
		bad = self.photons > (self.median() + threshold*self.sigma(robust=True)).reshape(shape)
		x, y, z = bad.nonzero()
		self.photons[x,y,z] = self.median()[x,y]
		self.speak('trimmed {0} cosmics from cube!'.format(np.sum(bad)))

	def master(self, which='photons'):
		'''
		The (calculated) master frame.
		'''
		return self.median(which)

	def nsigma(self, which='photons', robust=True):
		'''
		Calculate the number of sigma of each deviation from the median.
		'''
		array = self.__dict__[which].astype(np.float64)
		return (array - self.cubify(self.median(which)))/self.cubify(self.sigma(which, robust=robust))

	def write(self, normalization='none', directory='cube'):
		'''
		Save all the images to FITS, inside the cube directory.
		'''

		# make a directory for the normalization used
		dir = self.directory + normalization + '/'
		mkdir(dir)

		if normalization == 'none':
			flux = self.photons
		if normalization == 'nsigma':
			flux = self.nsigma()

		# loop through the images in the cube
		mkdir(directory)
		for i in range(self.n):
			# pick some kind of normalization for the image
			image = flux[i, :,:]

			filename = os.path.join(directory, normalization + '_{0:05.0f}.fits'.format(i))
			hdu = astropy.io.fits.PrimaryHDU(image)
			hdu.writeto(filename, overwrite=True)



			self.ccd.writeToFITS(image, filename)


	def plot(self, ax=None, normalization='none', ylim=[None, None], color='gray', **kw):

		# choose how to normalize the lightcurves for plotting
		if normalization.lower() == 'none':
			normalizationarray = self.cadence
			ylabel = 'Photons/s'
		elif normalization.lower() == 'master':
			normalizationarray = self.master().reshape(1, self.ypixels, self.xpixels)
			ylabel='Relative Flux'
		elif normalization.lower() == 'median':
			normalizationarray = self.median().reshape(1, self.ypixels, self.xpixels)
			ylabel='Relative Flux'

		# create a relative light curve (dF/F, in most cases)
		photonsnormalized = self.photons/normalizationarray


		# set up a logarithmic color scale (going between 0 and 1)
		# (I think this could be done better with a Norm/cmap)
		def cmap(x):
			zero = np.min(np.log(self.master()*0.5))
			span = np.max(np.log(self.master())) - zero
			#normalized = (np.log(x) -  zero)/span
			normalized = (np.log(x) -  zero)/span
			return plt.matplotlib.cm.Blues_r(normalized)
			#return plt.matplotlib.cm.YlGn(normalized)

		# create a plot
		scale = 1.5

		# set up the axes, if they don't already exist
		if ax is None:
			plt.figure('{} | normalization={}'.format(self, normalization),
						figsize = (self.xpixels*scale,self.ypixels*scale), dpi=72)
			gs = plt.matplotlib.gridspec.GridSpec(self.ypixels, self.xpixels, wspace=0, hspace=0)
			ax = {}
			a = None

			# loop over pixels (in x and y directions)
			for i in range(self.ypixels):
				for j in range(self.xpixels):

					# set up axis sharing, so zooming on one plot zooms the other
					ax[(i,j)] = plt.subplot(gs[-(i+1),j], sharex=a, sharey=a)
					a = ax[(i,j)]


					# color the plot panel based on the pixel's intensity
					a.patch.set_facecolor(cmap(self.median()[i,j]))

					# mess with the labels
					if i == 0 and j == 0:
						plt.setp(a.get_xticklabels(), rotation=90)
						a.set_xlabel('Time (s)')
						a.set_ylabel(ylabel)
					else:
						plt.setp(a.get_xticklabels(), visible=False)
						plt.setp(a.get_yticklabels(), visible=False)

		# loop over pixels (in x and y directions)
		for i in range(self.ypixels):
			for j in range(self.xpixels):
				ax[(i,j)].plot(self.time, photonsnormalized[:,i,j], color=color, **kw)

		# set the ylimits, if available
		#if ylim is None:
		#	ax.set_ylim(np.min(photonsnormalized), np.max(photonsnormalized))
		#else:
		ax[(i,j)].set_ylim(*ylim)

		return ax



from .ZoomFrame import *
from astropy.nddata.utils import Cutout2D
from matplotlib.patches import Rectangle


class LocalZoomFrame(ZoomFrame):
    frametype = 'Zoom'

    def __init__(self, source, position=(0, 0), size=(10, 10), zoom=10, **kwargs):
        '''
        Initialize a ZoomFrame that can display
        images taken from a zoomed subset of
        a `source` imshowFrame. This is kind
        of like a magnifying class you can
        drop down on an image.

        It *doesn't* add a new axis.

        Parameters
        ----------

        source : imshowFrame
                A frame into which this one will be zooming.

        position : (tuple)
                The (x, y) = (col, row) position of the center of zoom.
                (These are in original image coordinates,
                not those that have been transformed for display.)

        shape : (tuple)
                The (x, y) = (ncols, nrows) shape of the zoom.
        '''

        ZoomFrame.__init__(self,  **kwargs)

        #
        self.source = source
        self.source.includes.append(self)

        self.position = position
        self.size = size
        self.zoom = zoom

        self.titlefordisplay = '{} | {}'.format(self.frametype, self.position)

    def _get_times(self):
        '''
        Get the available times associated with this frame.
        '''
        return self.source._get_times()

    def _find_timestep(self, time):
        '''
        Get the timestep, by using the source frame.
        '''
        return self.source._find_timestep(time)

    def _get_image(self, time=None):
        '''
        Get the image at a given time (defaulting to the first time),
        by pulling it from the source frame.
        '''

        # get the image, transform to the rotated camera frame
        transformedbigimage, actual_time = self.source._get_image(time)

        # get the position of this stamp, transformed to the rotated camera frame
        transformedxy = self.source._transformxy(*self.position)
        self.cutout = Cutout2D(transformedbigimage,
                               transformedxy, self.size, mode='partial')

        # all coordinates for the cutout are now in the transformed coordinates
        cutoutimage = self.cutout.data
        return cutoutimage, actual_time

    def plot(self, time=None):
        '''
        Generate the (initial) plot for this local zoom frame.

        Individual features can be modified afterwards,
        through the .ax (the plotted axes) or the .plotted
        (dictionary of plotted elements) attributes.

        Parameters
        ----------

        time : astropy Time
            The time to plot, defaults to the first with None.
        '''

        # pull out the array to work on
        image, actual_time = self._get_image(time)

        # extract the cmap from the source image
        cmap, norm, ticks = self.source._cmap_norm_ticks(image)


        # these are in the rotated camera frame
        x, y = self.cutout.center_original
        ysize, xsize = self.cutout.shape

        left, right = x - xsize * self.zoom / 2, x + xsize * self.zoom / 2
        bottom, top = y - ysize * self.zoom / 2, y + ysize * self.zoom / 2

        extent = [left, right, bottom, top]

        zooms = [self.illustration.frames[k]
                 for k in self.illustration.frames.keys() if 'zoom' in k]
        zorder = zooms.index(self)

        self.plotted['image'] = self.source.ax.imshow(image,
                                                     extent=extent,
                                                     interpolation='nearest',
                                                     origin='lower',
                                                     norm=norm, cmap=cmap,
                                                     zorder=zorder)

        # reate a Rectangle patch
        rect = Rectangle((left, bottom), xsize * self.zoom, ysize * self.zoom,
                         linewidth=1, edgecolor='black', facecolor='none',
                         zorder=zorder + 0.5, clip_on=True)

        # Add the patch to the Axes
        self.plotted['boxonzoom'] = self.source.ax.add_patch(rect)

        self.source.ax.set_clip_on(True)
        # add a box to the source image (cutout must have ben created, if plot has happened)
        #boxonoriginal = self.cutout.plot_on_original(ax=self.source.ax, clip_on=True)

        try:
            timestep = self._find_timestep(time)
        except:
            timestep = None
        self.currenttimestep = timestep

    def update(self, time):
        '''
        Update this frame to a particular time (for use in animations).
        '''
        # update the data, if we need to
        timestep = self._find_timestep(time)
        image, actual_time = self._get_image(time)
        if image is None:
            return

        if timestep != self.currenttimestep:
            self.plotted['image'].set_data(image)

            if 'time' in self.plotted:
                timestring = self.source._timestring(actual_time)
                self.source.plotted['time'].set_text(timestring)
        self.currenttimestep = timestep
